/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 * its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

#define MAIN 1
#define TRUE 1
#define FALSE 0

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include <io_lib/Read.h>
#include <trace_align.h>
#include <sequtils.h>
#include <cl.h>
#include <readline.h>
#include <rescale16bit.h>
#include <sw.h>
#include <seed.h>
#include "svec_clip.h"

static const char *cvsid = "$Id: svec_clip.c 35448 2010-07-01 15:08:05Z jkb $";

/**
 * Enable verbose output from this utility. The greater the value the greater
 * the verbosity. Set to 0 for no additional diagnostic output.
 */
int verbosity = -1;

int
main(int argc, char **argv)
{
    int sp = 0, sc = 0;
    int shuffle = 20;
    int max_clip = 70;
    FILE *query_fp = NULL;
    char queryfile[MAX_FILENAME_LEN];
    SEQUENCE *query = NULL;
    char scf_file[MAX_FILENAME_LEN], trace_file[MAX_FILENAME_LEN];
    Scf *scf = NULL;
    Read *trace = NULL;

    FILE *fofn_fp = NULL;
    char fofn_file[MAX_FILENAME_LEN];	/* file of file names (SCF files) */

    getint("-verbose", &verbosity, argc, argv);

    getint("-sc", &sc, argc, argv);	/* cloning site in vector */
    getint("-sp", &sp, argc, argv);	/* primer offset in vector */
    getint("-shuffle", &shuffle, argc, argv);	/* no of shuffles */

    if (!sp) {
	fprintf(stderr, "cloning site sc not defined\n");
	exit(1);
    }
    if (!sc) {
	fprintf(stderr, "primer site sp not defined\n");
	exit(1);
    }
    getint("-max", &max_clip, argc, argv);

    if ((query_fp = argfile("-query", "r", argc, argv, queryfile)) != NULL) {
	query = get_fasta_sq(query_fp);
    }
    if ((fofn_fp = argfile("-fofn", "r", argc, argv, fofn_file)) != NULL) {

        /*
         * Batch mode using a file of file names, read_reading() will
         * deal with multiple file types.
         */

	if (sp && sc && query) {
	    while (tts_read_line(fofn_fp, trace_file) != EOF) {

                if((trace = read_reading(trace_file, TT_ANY)) != NULL) {

		    if((scf = read2scf(trace)) != NULL) {

                        if(scf->header.sample_size == 2) {
                            /* Rescale 16 bit traces for auto-edit */
                            rescale16bit(scf);
                        }

		        doSequenceAndTraceAlignment(query, scf, trace_file,
                            sc, sp, max_clip, shuffle, argc, argv);
                        scf_deallocate(scf);
		    } else {
                        fprintf(stderr, "Error: Can't open trace file: %s .\n",
                            scf_file);
                    }

                    read_deallocate(trace);
                }
	    }
	    return 0;
	}
    } else if (getarg("-scf", scf_file, argc, argv)) {

        /*
         * Single shot mode with a single scf file. Included here for backward
         * compatability. (DEPRECATED)
         */

	if((scf = read_scf(scf_file)) != NULL) {

            if(scf->header.sample_size == 2) {
                /* Rescale 16 bit traces for auto-edit */
                rescale16bit(scf);
            }

	    if (sp && sc && query) {
	        doSequenceAndTraceAlignment(query, scf, scf_file, sc, sp,
	            max_clip, shuffle, argc, argv);
                scf_deallocate(scf);
	        return 0;
            }
	}

    } else if (getarg("-trace", trace_file, argc, argv)) {
        /*
         * Single shot mode taking any of the file types supported by the
         * Staden IO lib.
         */

        if((trace = read_reading(trace_file, TT_ANY)) != NULL) {

	    if((scf = read2scf(trace)) != NULL) {

                if(scf->header.sample_size == 2) {
                    /* Rescale 16 bit traces for auto-edit */
                    rescale16bit(scf);
                }
	        if (sp && sc && query) {
	            doSequenceAndTraceAlignment(query, scf, trace_file, sc, sp,
	                max_clip, shuffle, argc, argv);
                    scf_deallocate(scf);
                    read_deallocate(trace);
	            return 0;
	        }
            }
        }
    }

    if(query)
        free_seq(query);

    /* Fall through means error */
    return 1;
}

/**
 * Perform the actual aluignment and report the results iof the match of the
 * query sequence to both the called sequence and (if necessary) the trace
 * data.
 *
 * @param query is a pointer to a SEQUENCE structure containing, typically,
 *     the vectore sequence that should be searched for.
 * @param scf is the pointer to the Scf data structure containing the data to
 *     be analysed.
 * @param scf_file is the string containing the name of the scf file currently
 *     under investigation.
 * @param sc is the cloning site.
 * @param sp is the primer offset.
 * @param max_clip
 * @param shuffle is the number of random permutations of the vector sequence
 *     that should be performed when assessing the reliability of a match.
 * @param argc is the count of the number of command line arguments
 * @param argv is the vector of command line arguments.
 */
static void
doSequenceAndTraceAlignment(SEQUENCE * query,
			    Scf * scf,
			    char *scf_file,
			    int sc,
			    int sp,
			    int max_clip,
			    int shuffle,
			    int argc,
			    char **argv)
{
    SEQUENCE *seq, *vec_3=NULL, *vec_5=NULL;
    int force_trace_align = FALSE;	/* Forces trace alignment in SCF files.
					   Disables sequence alignment for these
					   files. Has no effect on
					   Fasta files. */
    int clip_vector, clip_seq;
    int score, matches;
    float mean, stdev, var;
    int ok;
    int found_3 = FALSE, found_5 = FALSE;

    seq = scf2seq(scf, scf_file);	/* convert a scf  to a SEQUENCE */

    /* (a) 5' end */

    vec_5 = get_5_vec(query, sc, sp);

    getboolean("-forcetracealign", &force_trace_align, argc, argv);

    /*
     * Default is to perform this unless the user has specificaly 
     * chosen to force trace alignment.
     */
    if (!force_trace_align) {
	/*
	 * Try needleman sequence alignment. If this fails to get a sufficient
	 * number of matching bases then a shuffling procedure is called to 
	 * assess the quality of the match. In these cases mean, stdev will
	 * have meaningfull contents after the call. If the match was
	 * successful they will not. The small values added to the stdev
	 * are to 'avoid' divide by zero errors.
	 */

	try_seq_alignment(seq, vec_5, max_clip, argc, argv, &clip_seq,
			  &clip_vector, &score, &matches, &mean, &stdev, &ok,
			  &shuffle);
    }
    /* found the vector */
    if (!force_trace_align && ok && ok >= shuffle) {

	char *t;

	printf("SL %-15s %5d %-15s %5d %5d %5d   ", vec_5->name,
	       clip_vector, seq->name, clip_seq, score, matches);
	printf("    %5.2f %5.2f %6.3f   %2d / %2d  ", mean, stdev,
	       (score - mean) / (stdev + 1.0e-10), ok, shuffle);
	t = seq->s;
	while (t < &seq->s[clip_seq])
	    putc(*(t++), stdout);
	putc(' ', stdout);
	printf("\n");
	found_5 = TRUE;
    } else {			/* try the trace alignment */

	if (!force_trace_align) {
	    printf("! SL %-15s %5d %-15s %5d %5d %5d   ", vec_5->name,
		   clip_vector, seq->name, clip_seq, score, matches);
	    printf("    %5.2f %5.2f %6.3f   %2d / %2d  ", mean, stdev,
		   (score - mean) / (stdev + 1.0e-10), ok, shuffle);
	}
	try_trace_alignment(scf, vec_5, argc, argv, &clip_seq, &score,
			    &mean, &var, &ok, &shuffle);
	printf("SL %-15s %5d %-15s %5d %5d %5d ", vec_5->name,
	       vec_5->len, scf_file, clip_seq, score, 0);
	printf("    %5.2f %5.2f %6.3f   %2d / %2d trace aligned \n",
	       mean, var, (score - mean) / (var + 1.0e-10), ok,
	       shuffle);
    }


    if (!force_trace_align) {
	/* (b) 3' end */
	/* The 3' end is treated differently, mostly for reasons of
	   efficiency. We just make one sequence alignment (with NO
	   shuffles). THis should be ok because if sequencing vector
	   is present at the 3'end it will be obvious */

	vec_3 = get_3_vec(query, sc, seq->len);

	complement_seq(seq->s);

	shuffle = 0;
	try_seq_alignment(seq, vec_3, seq->len, argc, argv, &clip_seq,
			  &clip_vector, &score, &matches, &mean,
			  &stdev, &ok, &shuffle);

	if (ok && ok >= shuffle) {	/* found the vector */

	    char *t;

	    printf("SR %-15s %5d %-15s %5d %5d %5d   ", vec_3->name,
		   vec_3->len - clip_vector, seq->name,
		   seq->len - clip_seq + 1, score, matches);
	    printf("    %5.2f %5.2f %6.3f   %2d / %2d  ", mean, stdev,
		   (score - mean) / (stdev + 1.0e-10), ok, shuffle);
	    t = seq->s;
	    while (t < &seq->s[clip_seq])
		putc(*(t++), stdout);
	    putc(' ', stdout);
	    printf("\n");
	    found_3 = TRUE;
	}
	/* neither end found - test for completely vector */

	if (!found_3 && !found_5) {
	    int match = 3, mismatch = -5, gap = -5, neutral = 0;

	    getint("-match=%d", &match, argc, argv);
	    getint("-mismatch=%d", &mismatch, argc, argv);
	    getint("-gap=%d", &gap, argc, argv);

	    if (mismatch > 0)
		mismatch = -mismatch;
	    if (gap > 0)
		gap = -gap;

	    score =
		smith_waterman_score2(query->s, seq->s, query->len,
				      seq->len, match, mismatch, gap,
				      neutral, 0, -1, &clip_vector,
				      &clip_seq);
	    /*
	     * completely vector (match extands > 30% of the
	     * total read length)
	     */
	    if (score > 0.3 * match * seq->len) {
		printf("CC Completely Sequencing vector (svec_clip)\n");
		printf("PS Completely Sequencing vector\n");
	    }
	}
    }
    if(vec_3)
        free_seq(vec_3);
    if(vec_5)
        free_seq(vec_5);
    free_seq(seq);
}

/**
 * Fixup a vector sequence to get the 5' sequence for matching to a reading.
 * The process involves the following steps: Create a new SEQUENCE object
 * and duplicate the name and description fields from the supplied sequence.
 *
 * If the sp value is > 0 (i.e. a forward primer) then start copying from
 * the supplied sequence string starting at location 'sc' and wrapping round
 * to the start if you reach the end of the sequence. A base compliment is
 * then applied (n.b. this also reverses the sequence string.)
 *
 * If the sp value is <=0 (i.e. a reverse primer) 
 *
 * @param seq is the original (full vector sequence)
 * @param sc is the cloning site in the sequencing vector
 * @param sp is the primer offset from the cloning site 
 *
 * @returns a SEQUENCE object representing the ...
 */
SEQUENCE *
get_5_vec(SEQUENCE * seq, int sc, int sp)
{

    SEQUENCE *vec = (SEQUENCE *) calloc(1, sizeof(SEQUENCE));
    int n, len;

    vec->name = strdup(seq->name);
    if (sp > 0) {		/* forward primer */
	vec->s = (char *) calloc(sp + 2, sizeof(char));
	len = strlen(seq->s);
	for (n = 0; n <= sp; n++)
	    vec->s[n] = seq->s[(sc + n) % len];
	vec->len = strlen(vec->s);
	complement_seq(vec->s);
	if (verbosity > 1)
	    printf("! vec %s\n", vec->s);
	return vec;
    } else {			/* reverse primer */
	sp = -sp;
	vec->s = (char *) calloc(sp + 2, sizeof(char));
	len = strlen(seq->s);
	for (n = 0; n <= sp; n++)
	    vec->s[n] = seq->s[(sc - sp + n + len) % len];
	vec->len = strlen(vec->s);
	if (verbosity > 1)
	    printf("! vec %s\n", vec->s);
	return vec;
    }
}

SEQUENCE *
get_3_vec(SEQUENCE * seq, int sc, int Len)
{
/* fixup a vector sequence to get the 3' sequence for matching to a reading 

   sc is the cloning site in the sequencing vector
   Len is the max length to align

 */

    SEQUENCE *vec = (SEQUENCE *) calloc(1, sizeof(SEQUENCE));
    int n, len;

    vec->name = strdup(seq->name);
    len = strlen(seq->s);
    if (Len > 0) {		/* forward primer */
	if (Len > len)
	    Len = len;
	vec->s = (char *) calloc(Len + 1, sizeof(char));
	for (n = 0; n < Len; n++)
	    vec->s[Len - n - 1] = seq->s[(len + sc - n) % len];
	vec->len = strlen(vec->s);
/*      complement_seq(vec->s);    */
	if (verbosity > 1)
	    printf("! 3' %d %d vec %s\n", Len, len, vec->s);
    } else {			/* reverse primer */
	Len = -Len;
	if (Len > len)
	    Len = len;
	vec->s = (char *) calloc(Len + 1, sizeof(char));
	len = strlen(seq->s);
	for (n = 0; n < Len; n++)
	    vec->s[n] = seq->s[(sc + n + len) % len];
	vec->len = strlen(vec->s);
	complement_seq(vec->s);
/*      if(verbosity > 1)
   printf("! vec %d %s\n", Len, vec->s); */
    }
    return vec;

}

/**
 * Returns the Needleman Wunsch score for aligning the end of query to
 * the begining of seq matches score +match, mismatches +mismatch,
 * gaps +gap (no gap start penalty)
 *
 * @param query is the SEQUENCE representing the string to be matched in
 *     the larger 'seq' sequence.
 * @param seq is the larger sequence object that should be scanned for a match.
 * @param match is the score (to be used in the substitution matrix where
 *     a complete match occurs).
 * @param mismatch is the score (to be used in the substitution matrix where
 *     a complete mismatch occurs).
 * @param gap is the penalty (to be used in the substitution matrix where
 *     a gap must be introduced).
 * @param max is the maximum offset into the 'seq' object that should be
 *     explored during the matching process.
 * @param align_mode is a flag, the value drawn from the ENUMERATION of 
 *     different alignment algotithms, which indicates whether a Needleman
 *     global alignment should be conducted or whether a Smith-Waterman
 *     optimal local alignment should be carried out.
 * @param q a return parameter indicating the position in the query sequence
 *     where the match started.
 * @param s a return parameter indicating the position in the sequence string
 *     where the match started.
 * @param score is a return paramater holding the maximum match score found.
 * @param matches is a return parameter holding the count of matching bases.
 */
void
end_align(SEQUENCE * query, SEQUENCE * seq,
	  int match, int mismatch, int gap, int max, int align_mode,
	  int *q, int *s, int *score, int *matches)
{
    int *A, *B, *a, *b, *c;
    int *countA, *countB, *counta, *countb;
    int i1, i2;
    int max_score = -1000000;	/* arbitrary large negative value */
    int c1 = 0, c2 = 0;
    int m;
    int d, u, l;
    int start1, stop1;
    int start2, stop2;
    int x2;
    char *lower;
    static int mat_init_ok = 0;
    int match_count = 0;
    char pad_char = '*';

    /*
     * Build a similarity matrix based on the supplied (scalar) values for a
     * match, mismatch, gap etc. This is a single shot operation that initalises
     * the sim_mat 2D array.
     */
    if (!mat_init_ok) {
	mat_init(match, mismatch, gap, 0, pad_char);
	mat_init_ok = 1;
    }
    /*
     * NB. Query is the 'short' 40'ish long string
     */
    A = (int *) calloc(query->len + 1, sizeof(int));
    B = (int *) calloc(query->len + 1, sizeof(int));

    /*
     * NB. calloc'd arrays are pre-initalised to have zero value.
     */
    countA = (int *) calloc(query->len + 1, sizeof(int));
    countB = (int *) calloc(query->len + 1, sizeof(int));

    lower = (char *) downcase(strdup(query->s));

    /*
     * Array based at the first offset of the original. a[-1]->A[0]
     */
    a = A + 1;
    b = B + 1;

    counta = countA + 1;
    countb = countB + 1;

    /*
     * Indicies of the first and last elements of the 'long' sequence string.
     */
    start2 = 0;
    stop2 = seq->len - 1;


    /*
     * If the user has supplied a positive max parameter and its less than
     * the length of the sequence then we use that value as the maximum
     * extent of the search through the sequence.
     */
    if (max > 0 && stop2 > max)
	stop2 = max;

    /*
     * Itterate over the sequence stopping at the  end of the sequence or after
     * 'max' characters which ever occurs first.
     */
    for (i2 = start2; i2 <= stop2; i2++) {

	/*
	 * a - Previous row of Matrix. (a[-1] = 0, by use of calloc A)
	 * b - Current row of the Matrix. (b[-1] = 0, by use of calloc B)
	 */

	c = a;
	a = b;
	b = c;

	c = counta;
	counta = countb;
	countb = c;

	start1 = 0;
	stop1 = query->len - 1;

	x2 = tolower(seq->s[i2]);

	/*
	 * Space conserving optimisation which only maintains the current
	 * row of the 'F' matrix and the preceding row. (This is all that
	 * is required to compute all the values on a row by row basis.
	 * On each itteration of the outer loop these are swapped over and
	 * the new values overwrite those stored in 'b'.
	 */

	/*
	 * Itterate over the query string.
	 */

	for (i1 = start1; i1 <= stop1; i1++) {

	    /*
	     * Calculate the terms of the MAX function.
	     * d - diagonal term.
	     * u - up.
	     * l - left.
	     */

	    d = a[i1 - 1] + (m = sim_mat[lower[i1]][x2]);
	    u = a[i1] + gap;
	    l = b[i1 - 1] + gap;

	    /*
	     * MAX of the possibilities is chosen.
	     */
	    if (d > u) {
		if (d > l) {
		    /* d holds the MAX value */
		    b[i1] = d;
		    countb[i1] = (m > 0) + counta[i1 - 1];
		} else {
		    /* l holds the MAX value */
		    b[i1] = l;
		    countb[i1] = countb[i1 - 1];
		}
	    } else if (u > l) {
		/* u holds the MAX value */
		b[i1] = u;
		countb[i1] = counta[i1];
	    } else {
		/* l holds the MAX value */
		b[i1] = l;
		countb[i1] = countb[i1 - 1];
	    }

	    /*
	     * Additional term for the MAX operator in SW. This sets the
	     * value to zero if the current best match is negative (and
	     * we are using smith-waterman.
	     */
	    if (align_mode == SMITH_WATERMAN && b[i1] < 0) {	/* SW */
		b[i1] = 0;	/* SW */
		countb[i1] = 0;
	    }
	}

	i1--;

	/*
	 * If the value at the end of the match is > than any found so far
	 * then keep it, the count of the number of matches and the current
	 * indicies for the loops. i2 - the position in the main sequence
	 * string and  i1 will be the last offset in the pattern string.
	 *
	 * There is a special case for the first pass where the current
	 * implementation uses a large, arbitary, negative value (-100000).
	 * this should ensure that this will be set to the first sensible
	 * value it finds.
	 */
	if (b[i1] > max_score) {
	    max_score = b[i1];
	    match_count = countb[i1];
	    c1 = i1;
	    c2 = i2;
	}
    }

    free(A);
    free(B);
    free(countA);
    free(countB);
    free(lower);

    /*
     * Output parameter mapping.
     */
    *q = c1;
    *s = c2;
    *score = max_score;
    *matches = match_count;
}

void
try_trace_alignment(Scf * scf, SEQUENCE * vec, int argc, char **argv,
		    int *clip_point, int *score, float *mean, float *var,
		    int *ok, int *shuffle)
{

    Header *h;
    int bases;
    int trace_start, trace_end;
    int *coord, *w_score, *w_max, *t_score, *t_max;
    int base_start, base_end, t_start, t_end;
    int s;
    float floor = .90;
    float t_weight = 2;
    int align_mode = NEEDLEMAN;
    char mode[256];
    int max = 70;
    int min_trace = 7;
    int seed;
    int n, s_score;
    SEQUENCE *shuffled_vec = NULL;
    FILE *ps;
    char ps_file[256];
    int window_band = 0;
    t_params params;
    int w, wm;
    int rescore = 0;

    getint("-max=%d", &max, argc, argv);	/* max clip in sequence */

    if (*shuffle > 0) {
	getseed(&seed, argc, argv);
	shuffled_vec = seqdup(vec);
    }
    getint("-min=%d", &min_trace, argc, argv);
    getfloat("-floor=%f", &floor, argc, argv);
    getfloat("t_weight=%f", &t_weight, argc, argv);

    strcpy(mode, "needleman");
    if (getarg("-mode=%s", mode, argc, argv))
	legal_string(mode, align_modes, 2, &align_mode);

    params.minstate = min_trace;
    params.floor = floor;
    params.deriv_weight = t_weight;
    params.mode = align_mode;
    params.samples_per_base = window_band;
    params.gap_penalty = 0;

    h = &scf->header;
    bases = vec->len;
    trace_start = 0;
    if (h->bases < max)
	trace_end = h->samples;
    else
	trace_end = scf->bases[max - 1].peak_index;

    coord = (int *) calloc(bases + 1, sizeof(int));
    w_score = (int *) calloc(bases + 1, sizeof(int));
    w_max = (int *) calloc(bases + 1, sizeof(int));
    t_score = (int *) calloc(bases + 1, sizeof(int));
    t_max = (int *) calloc(bases + 1, sizeof(int));

    trace_align(scf, trace_start, trace_end, vec->s, 0, vec->len - 1, &params,
		rescore, coord, w_score, score, &base_start, &base_end,
		&t_start, &t_end);

    if (ps = argfile("-ps=%s", "w", argc, argv, ps_file)) {
	ps_print_alignment(ps, scf, trace_start, trace_end,
			   base_end - base_start + 1, vec->s, base_start,
			   base_end, coord, ps_file);
	fclose(ps);
    }
    for (s = 0; s < h->bases; s++)
	if (scf->bases[s].peak_index > t_end)
	    break;

    *clip_point = s;


    *ok = 0;
    *mean = *var = 0.0;

    wm = 0;
    w = 0;
    for (s = base_start; s <= base_end; s++) {
	wm += w_max[s];
	w += w_score[s];
    }
    printf("! len: %d, w: %d wm %d  %.3f\n", base_end - base_start + 1, w, wm,
	   0.90 * wm);
    if (base_end - base_start + 1 > 30 && w > 0.90 * wm)	/* very strong similarity - no need to shuffle */
	*ok = *shuffle;
/* very weak - obviously no good */
/*  else if ( w < 0.5 *wm ) 
   *ok = 0; */
    else if (*shuffle) {	/* do shuffles to test */
	for (n = 0; n < *shuffle; n++) {
	    shuffle_seq(shuffled_vec, 1, &seed);
	    trace_align(scf, trace_start, trace_end, shuffled_vec->s, 0,
		      shuffled_vec->len - 1, &params, rescore, coord, w_score,
			&s_score, &base_start, &base_end, &t_start, &t_end);

	    *mean += (float) s_score;
	    *var += (float) s_score * (float) s_score;
	    *ok += (*score > s_score);
	}
	*mean /= *shuffle;
	*var =
	    (*var - *shuffle * (*mean) * (*mean)) / (*shuffle - 1.0 + 1.0e-5);
	*var = sqrt(*var) + 1.0e-5;
    }

    if(shuffled_vec)
        free_seq(shuffled_vec);

    free(coord);
    free(w_score);
    free(w_max);
    free(t_score);
    free(t_max);
}

/**
 * Try to match the vector to a sequence using a neddleman-type
 * alignment.  If a match of over 75% ot the bases (or to whatever
 * match_fraction is set to) is found then this is 
 * treated as definitely correct.  Otherwise the alignment is
 * tested by repeated shuffling of the vector sequence. Simple statistics
 * for how random shuffles of the vector sequence performed on the same
 * data are returned in mean and stdev.
 *
 * The sequences should be input in such a way that we are always
 * matching the end of the vector to the start of the query.  This is
 * natural for 5' end, but you need to reverse-complement for the 3' end
 *
 * @param seq
 * @param vec
 * @param max_clip
 * @param argc
 * @param argv
 * @param clip_pos
 * @param clip_pos_vec
 * @param score
 * @param matches
 * @param mean is the mean needleman match score (if the shuffling
 *        procedure is carried out) for the set of random base shuffles
 *        in the target vector sequence against the original target data.
 * @param stdev is the standard dev. of needleman match scores (if the shuffling
 *        procedure is carried out) for the set of random base shuffles
 *        in the target vector sequence against the original target data.
 * @param ok
 * @param shuffle is the number of shuffles that should be carried out if the
 *        match_fraction acceptance criteria is not satisfied.
 *
 */

void
try_seq_alignment(SEQUENCE * seq, SEQUENCE * vec,
		  int max_clip,
		  int argc, char **argv, int *clip_pos,
		  int *clip_pos_vec, int *score, int *matches,
		  float *mean, float *stdev, int *ok, int *shuffle)
{
    SEQUENCE *shuffled_vec = NULL;
    int match = 3, mismatch = -5, gap = -5;
    int n;

    /*
     * Default fraction of bases that need to be matched to have confidence
     * that the match is 'real'.
     */
    float match_fraction = 0.75;

    int align_mode = NEEDLEMAN;
    char mode[256];
    int seed;
    int q, s;
    int s_score, ss, s_matches;
    getint("-match=%d", &match, argc, argv);
    getint("-mismatch=%d", &mismatch, argc, argv);
    getint("-gap=%d", &gap, argc, argv);
    getfloat("-matchfrac=%f", &match_fraction, argc, argv);

    if (mismatch > 0)
	mismatch = -mismatch;
    if (gap > 0)
	gap = -gap;

    if (*shuffle > 0) {
	getseed(&seed, argc, argv);
	shuffled_vec = seqdup(vec);
    }
    strcpy(mode, "needleman");
    if (getarg("-mode=%s", mode, argc, argv))
	legal_string(mode, align_modes, 2, &align_mode);

    *mean = 0.0;
    *stdev = 0.0;
    *ok = 0;

    if (verbosity > 1) {
	printf("!seq: %s\n", seq->s);
	printf("!vec: %s\n", vec->s);
    }
    end_align(vec, seq, match, mismatch, gap, max_clip, align_mode, &q, &s,
	      score, matches);
    printf("! score %d matches %d\n", *score, *matches);
    *clip_pos = s + 1;
    *clip_pos_vec = q + 1;

    /*
     * An alignment with > match_fraction should be correct, so don't
     * bother with the validation process using shuffling.
     */
    if (*matches > match_fraction * strlen(vec->s))
	*ok = (*shuffle ? *shuffle : 1);
    else if (*shuffle) {
	if (verbosity > 0) {
	    printf("! shuffling %d\n", *shuffle);
	}
	/*
	 * Single pass, slightly unclear, implementation of the mean and
	 * variance calculation. Based on var=((sum(x^2)-1/n*(sum(x)^2))/n-1
	 * small values have been added to avoid issues with divide by zero
	 * errors. Arguable that this is NOT the way to deal with this issue
	 * since it should avoid such situations not just paper over the cracks.
	 *
	 * Corrected to reflect the fact that the old return parameter 'var'
	 * returned teh standard deviation and not the variance as suggested.
	 */
	printf("! shuffling %d\n", *shuffle);
	for (n = 0; n < *shuffle; n++) {
	    shuffle_seq(shuffled_vec, 1, &seed);
	    end_align(shuffled_vec, seq, match, mismatch, gap, max_clip,
		      align_mode, &q, &ss, &s_score, &s_matches);
/*        printf("! %5d score %d\n", n, s_score ); */
	    (*mean) += s_score;
	    (*stdev) += s_score * s_score;
	    *ok += ((*score) > s_score);
	}
	(*mean) /= *shuffle;
	(*stdev) = ((*stdev) - *shuffle * (*mean) * (*mean)) / (*shuffle - 1.0 + 1.0e-5);
	(*stdev) = sqrt((*stdev)) + 1.0e-5;
    }

    if(shuffled_vec)
        free_seq(shuffled_vec);
}

SEQUENCE *
scf2seq(Scf * scf, char *name)
{				/* extract the sequence from the scf */
    Header *h = &scf->header;
    SEQUENCE *seq = (SEQUENCE *) calloc(1, sizeof(SEQUENCE));
    int i;

    seq->len = h->bases;
    seq->s = (char *) calloc(seq->len + 1, sizeof(char));
    for (i = 0; i < seq->len; i++)
	seq->s[i] = scf->bases[i].base;
    seq->name = (char *) strdup(name);

    return seq;
}
