/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 * its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include <io_lib/Read.h>
#include <sequtils.h>
#include <cl.h>
#include <tracealign.h>
#include <trace_align.h>
#include <rescale16bit.h>

static const char *cvsid = "$Id: tracealign.c 35448 2010-07-01 15:08:05Z jkb $";

/**
 * Enable verbose output from this utility. The greater the value the greater
 * the verbosity. Set to 0 for no additional diagnostic output.
 */
int verbosity = -1;

int
main(int argc, char **argv)
{
    FILE *query_fp = NULL;
    char queryfile[MAX_FILENAME_LEN];
    SEQUENCE *query = NULL;
    char trace_file[MAX_FILENAME_LEN];
    Read *trace = NULL;

    FILE *fofn_fp = NULL;
    char fofn_file[MAX_FILENAME_LEN];	/* file of file names (SCF files) */

    getint("-verbose", &verbosity, argc, argv);

    if ((query_fp = argfile("-query", "r", argc, argv, queryfile)) != NULL) {
	query = get_fasta_sq(query_fp);
    }

    if (getarg("-trace", trace_file, argc, argv)) {

	/*
         * Single shot mode taking any of the file types supported by the
	 * Staden IO lib.
	 */

	if((trace = read_reading(trace_file, TT_ANY)) != NULL) {

            Scf *scf;

            if((scf = read2scf(trace)) != NULL) {

                if(scf->header.sample_size == 2)
                    rescale16bit(scf);

                if (query) {
		    doSequenceToTraceAlignment(query, scf, trace_file,
                        argc, argv);
                    scf_deallocate(scf);
                    read_deallocate(trace);
		    return 0;
	        }
            }
        }
    }

    if(query)
        free_seq(query);

    /* Fall through means error */
    return 1;
}

/**
 * Perform the actual aluignment and report the results iof the match of the
 * query sequence to both the called sequence and (if necessary) the trace
 * data.
 *
 * @param query is a pointer to a SEQUENCE structure containing, typically,
 *     the vectore sequence that should be searched for.
 * @param scf is the pointer to the Scf data structure containing the data to
 *     be analysed.
 * @param scf_file is the string containing the name of the scf file currently
 *     under investigation.
 * @param argc is the count of the number of command line arguments
 * @param argv is the vector of command line arguments.
 */
static void
doSequenceToTraceAlignment(SEQUENCE * query,
			    Scf * scf,
			    char *scf_file,
			    int argc,
			    char **argv)
{
    SEQUENCE *seq;
    int clip_vector, clip_seq;
    int score;

    seq = scf2seq(scf, scf_file);	/* convert a scf  to a SEQUENCE */

    try_trace_alignment(scf, query, argc, argv, &clip_seq, &score);
    printf("SL %-15s %5d %-15s %5d %5d\n", query->name,
       query->len, scf_file, clip_seq, score);

    free_seq(seq);
}

void
try_trace_alignment(Scf * scf, SEQUENCE * vec, int argc, char **argv,
		    int *clip_point, int *score)
{

    Header *h;
    int bases;
    int trace_start, trace_end;
    int *coord, *w_score, *w_max, *t_score, *t_max;
    int base_start, base_end, t_start, t_end;
    int s;
    float floor = .90;
    float t_weight = 2;
    int align_mode = SMITH_WATERMAN;
    char mode[MAX_FILENAME_LEN];
    int max = 70;
    int min_trace = 7;
    int seed;
    int n, s_score;
    FILE *ps, *out;
    char ps_file[MAX_FILENAME_LEN], out_file[MAX_FILENAME_LEN];
    int window_band = 0;
    t_params params;
    int w, wm;
    int rescore = 0;

    getint("-max=%d", &max, argc, argv);	/* max clip in sequence */

    getint("-min=%d", &min_trace, argc, argv);
    getfloat("-floor=%f", &floor, argc, argv);
    getfloat("-t_weight=%f", &t_weight, argc, argv);

    strcpy(mode, "smithwaterman");
    if (getarg("-mode=%s", mode, argc, argv))
	legal_string(mode, align_modes, 2, &align_mode);

    params.minstate = min_trace;
    params.floor = floor;
    params.deriv_weight = t_weight;
    params.mode = align_mode;
    params.samples_per_base = window_band;
    params.gap_penalty = 0;

    h = &scf->header;
    bases = vec->len;
    trace_start = 0;
    if (h->bases < max)
	trace_end = h->samples;
    else
	trace_end = scf->bases[max - 1].peak_index;

    coord = (int *) calloc(bases + 1, sizeof(int));
    w_score = (int *) calloc(bases + 1, sizeof(int));
    w_max = (int *) calloc(bases + 1, sizeof(int));
    t_score = (int *) calloc(bases + 1, sizeof(int));
    t_max = (int *) calloc(bases + 1, sizeof(int));

    trace_align(scf, trace_start, trace_end, vec->s, 0, vec->len - 1, &params,
		rescore, coord, w_score, score, &base_start, &base_end,
		&t_start, &t_end);

    if (ps = argfile("-ps=%s", "w", argc, argv, ps_file)) {
	ps_print_alignment(ps, scf, trace_start, trace_end,
			   base_end - base_start + 1, vec->s, base_start,
			   base_end, coord, ps_file);
	fclose(ps);
    }
    if (out = argfile("-out=%s", "w", argc, argv, out_file)) {
	write_scf_alignment(out, scf, trace_start, trace_end,
			   base_end - base_start + 1, vec->s, base_start,
			   base_end, coord);
	fclose(out);
    }
    for (s = 0; s < h->bases; s++)
	if (scf->bases[s].peak_index > t_end)
	    break;

    *clip_point = s;

    wm = 0;
    w = 0;
    for (s = base_start; s <= base_end; s++) {
	wm += w_max[s];
	w += w_score[s];
    }
    printf("! len: %d, w: %d wm %d  %.3f\n", base_end - base_start + 1, w, wm,
	   0.90 * wm);

    free(coord);
    free(w_score);
    free(w_max);
    free(t_score);
    free(t_max);
}


/*
 * Perform a linear search for the maximum value of the trace data in the
 * selected channel in the range begin..end.
 *
 * FILE PRIVATE
 */
static uint_4
findPeakPosition(Scf *data, char channel, uint_4 begin, uint_4 end) {

    Header hdr;
    uint_4 size, i, maxval, maxpos, first=1;

    hdr = data->header;
    size = hdr.sample_size;

    if(size == 1) { /* 8 bit data */
        Samples1    *samples =  data->samples.samples1;
        maxpos = begin;
        for(i = begin; i < end; i++) {
            switch(channel) {
                case 'A':
                case 'a':
                         if(first) {
                             first = 0;
                             maxval = samples[i].sample_A;
                         } else {
                             if(samples[i].sample_A > maxval) {
                                 maxval = samples[i].sample_A;
                                 maxpos = i;
                             }
                         }
                         break;
                case 'C':
                case 'c':
                         if(first) {
                             first = 0;
                             maxval = samples[i].sample_C;
                         } else {
                             if(samples[i].sample_C > maxval) {
                                 maxval = samples[i].sample_C;
                                 maxpos = i;
                             }
                         }
                         break;
                case 'T':
                case 't':
                         if(first) {
                             first = 0;
                             maxval = samples[i].sample_T;
                         } else {
                             if(samples[i].sample_T > maxval) {
                                 maxval = samples[i].sample_T;
                                 maxpos = i;
                             }
                         }
                         break;
                case 'G':
                case 'g':
                         if(first) {
                             first = 0;
                             maxval = samples[i].sample_G;
                         } else {
                             if(samples[i].sample_G > maxval) {
                                 maxval = samples[i].sample_G;
                                 maxpos = i;
                             }
                         }
                         break;
            }
        }
        /* Have the maximum here */
        return maxpos;

    } else if(size == 2) { /* 16 bit data */

        Samples2    *samples =  data->samples.samples2;
        maxpos = begin;
        for(i = begin; i < end; i++) {
            switch(channel) {
                case 'A':
                case 'a':
                         if(first) {
                             first = 0;
                             maxval = samples[i].sample_A;
                         } else {
                             if(samples[i].sample_A > maxval) {
                                 maxval = samples[i].sample_A;
                                 maxpos = i;
                             }
                         }
                         break;
                case 'C':
                case 'c':
                         if(first) {
                             first = 0;
                             maxval = samples[i].sample_C;
                         } else {
                             if(samples[i].sample_C > maxval) {
                                 maxval = samples[i].sample_C;
                                 maxpos = i;
                             }
                         }
                         break;
                case 'T':
                case 't':
                         if(first) {
                             first = 0;
                             maxval = samples[i].sample_T;
                         } else {
                             if(samples[i].sample_T > maxval) {
                                 maxval = samples[i].sample_T;
                                 maxpos = i;
                             }
                         }
                         break;
                case 'G':
                case 'g':
                         if(first) {
                             first = 0;
                             maxval = samples[i].sample_G;
                         } else {
                             if(samples[i].sample_G > maxval) {
                                 maxval = samples[i].sample_G;
                                 maxpos = i;
                             }
                         }
                         break;
            }
        }
        /* Have maximum value here */
        return maxpos;
    } else {
        /* FAULT */
        fprintf(stderr, "Internal Error, invalid SCF data size\n");
        exit(1);
    }
}

/**
 * Writes a scf file to the supplied file stream. This file contains the
 * original trace data but has the base information replaced by a new set
 * of base positions which are computed from the trace alignment.
 *
 * <b>FILE PRIVATE</b>
 *
 * @param fp is the file stream to write the scf data into. This should be
 *    open for writing.
 * @param scf is the pointer to the original scf data structure. This is used
 *    to obtain the original trace and comment information.
 * @param trace_start is the coordinate in the trace data where the output
 *    trace should start.
 * @param trace_end is the coordinate in the trace data where the output
 *    trace should terminate.
 * @param base_count is the number of base codes that should be output.
 * @param base is the array of base codes (for the complete query sequence).
 * @param base_start is the index into the 'base' array that corresponds to
 *    the first aligning base in the trace.
 * @param base_end is the index into the 'base' array that corresponds to the
 *    that last aligning base in the trace.
 * @param trace_coord is the location of the start of a base to trace alignment
 *    segment. This is indexed from base_start..base_end and provides the 
 *    coordinate in the trace where the alignment algorithim has determined
 *    the start of the region associated with the base code found in the
 *    corresponding location in base.
 */
static void
write_scf_alignment (FILE *fp, Scf *scf, int trace_start, int trace_end,
                           int base_count, char *base, int base_start,
                           int base_end, int *trace_coord) {

    int    i, j;
    Scf    tmpScf, *newScf;
    Bases  *newBases = NULL;

    tmpScf = *scf; /* copy the contents of the supplied Scf structure */

    if (trace_coord) {

        newScf = &tmpScf;

        /* Calloc means all quality info = 0 */
        newBases = (Bases *)calloc((base_end-base_start), sizeof(Bases));

        i=0;
	for (j = base_start+1; j < base_end; j++) {
            newBases[i].base = base[j];
	    newBases[i++].peak_index = findPeakPosition(scf, base[j],
                                           trace_coord[j], trace_coord[j+1]);
        }

        (newScf->header).bases = (uint_4)i;
        newScf->bases = newBases;
        
    } else {
        newScf = scf; /* default does nothing to the scf data */
    }

    fwrite_scf(newScf, fp);

    if(newBases) 
        free(newBases);
}


SEQUENCE *
scf2seq(Scf * scf, char *name)
{				/* extract the sequence from the scf */
    Header *h = &scf->header;
    SEQUENCE *seq = (SEQUENCE *) calloc(1, sizeof(SEQUENCE));
    int i;

    seq->len = h->bases;
    seq->s = (char *) calloc(seq->len + 1, sizeof(char));
    for (i = 0; i < seq->len; i++)
	seq->s[i] = scf->bases[i].base;
    seq->name = (char *) strdup(name);

    return seq;
}
