/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

/**
 * @file
 * This file defines routines associated with manipulating 'databases' of
 * sequence data.
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "sequtils.h"
#include <readline.h>

static const char *cvsid = "$Id: seq_db.c 18816 2006-08-03 16:39:42Z rmd $";

static int database_count = 0;
static DATABASE *databases[MAX_DB_COUNT];

/**
 * Opens a 'database' of sequence data. This routine wraps a cache of 
 * previously opened databases (keyed on the name). If the name is found
 * in the cache then its handle is returned. Otherwise a cascade of 
 * specalised database opening routines is called (in the following order)
 * to open embl, nbrf and fasta files.
 *
 * @param name is the file name for the file containing the sequence data.
 *
 * @returns a database handle if the named database is either in the cache
 *     or it can be successfully opened. The routine forces the calling
 *     program to terminate if the database cannot be opened.
 */
DATABASE *
open_database(char *name)
{

    DATABASE *db;

    if ((db = which_database(name)) != 0)
	return db;

    if ((db = open_embl_database(name)) != 0)
	return db;
    else if ((db = open_nbrf_database(name)) != 0)
	return db;
    else if ((db = open_fasta_database(name)) != 0)
	return db;
    else {
	fprintf(stderr, "ERROR could not open database %s\n", name);
	exit(1);
    }
}

/**
 * Checks the cache of DATABASES to see if there is an entry with the 
 * matching name. This performs a linear search of the cache (currently
 * implemented as and array).
 *
 * @param name is the name of the database that should be searched for.
 *
 * @returns the database handle associated with the name or NULL if it
 *     cannot be found.
 */
DATABASE *
which_database(char *database_name)
{
    int n;

    for (n = 0; n < database_count; n++)
	if (!strcmp(database_name, databases[n]->database))
	    return databases[n];
    return NULL;
}

/**
 * Reads a set of sequence names from the supplied file stream. The names
 * are arranged one per line and are of the form required by the read_seq()
 * routine. The routine will try all of the sequence identifiers in the
 * file until it finds one which can be read successfully. Comments will
 * also be skipped (see tts_skip_comments()).
 *
 * @param list_file is the file stream containing the names of the sequence
 *     identifiers that should be read in.
 *
 * @returns 
 */
SEQUENCE *
next_seq_from_list_file(FILE *list_file)
{
    char line[256], seq[256];
    SEQUENCE *sq = NULL;

    *line = 0;

    while (!sq && tts_skip_comments(list_file, line) != EOF) {
	sscanf(line, "%s", seq);
	sq = read_seq(seq);
    }

    return sq;
}


/**
 * Opens a, EMBL format, sequence file. This builds an internal database
 * cache entry (which must be released by the user when no longer required)
 * that contains its name, type and the file handles that contain
 * the sequence and sequence-index data. This will return a NULL if any
 * of the files (data and index) cannot be opened.
 *
 * @see open_fasta_database, open_nbrf_database
 *
 * @see http://www.psc.edu/general/software/packages/seq-intro/emblfile.html
 *     for a description of this file format.
 *
 * @param name is the base-name of the file containing a EMBL format
 *     sequence database. The name will have '.index' and '.dat' appended
 *     and it expects to find both files available for reading.
 *
 * @returns a DATABASE handle if the file is successfully opened and NULL
 *     otherwise. 
 */
DATABASE *
open_embl_database(char *name)
{

    char buf[256];
    DATABASE *db;

    /* Make a db entry */
    db = (DATABASE *) calloc(1, sizeof(DATABASE));
    db->database = (char *) strdup(name);
    db->datafile = openfile_in_seqpath(name, "dat", "r", buf);
    db->indexfile = openfile_in_seqpath(name, "index", "r", buf);
    db->type = EMBL;

    if (!db->datafile || !db->indexfile) {
	free(db);
	return NULL;
    }

    /* Read the index info into a memory table. db is both input and output */
    make_embl_index(db);

    fprintf(stderr, "\n! embl format database %s opened: %d sequences\n", name,
	    db->sequences);

    /* deal with adding it to the cache of DATABASE entries. */
    add_database(db);

    return db;
}

/**
 * Opens a, NBRF format, sequence file. This builds an internal database
 * cache entry (which must be released by the user when no longer required)
 * that contains its name, type and the file handles that contain
 * the sequence and sequence-index  and ref data. This will return a
 * NULL if any of the files (data, index and ref) cannot be opened.
 *
 * @see open_fasta_database, open_embl_database
 *
 * @param name is the base-name of the file containing the sequence data.
 *     This routine assumes that the name will have '.seq', '.ref' and 
 *     '.index' appended to it and that each of these files can be opened
 *     for reading.
 *
 * @returns a DATABASE handle if the file is successfully opened and NULL
 *     otherwise. 
 */
DATABASE *
open_nbrf_database(char *name)
{
    char buf[256];
    DATABASE *db;

    /* Make a db entry */
    db = (DATABASE *) calloc(1, sizeof(DATABASE));
    db->database = (char *) strdup(name);
    db->datafile = openfile_in_seqpath(name, "seq", "r", buf);
    db->textfile = openfile_in_seqpath(name, "ref", "r", buf);
    db->indexfile = openfile_in_seqpath(name, "index", "r", buf);
    db->type = NBRF;

    if (!db->datafile || !db->textfile || !db->indexfile) {
	free(db);
	return NULL;
    }

    /* Read the index info into a memory table. db is both input and output */
    make_nbrf_index(db);

    fprintf(stderr, "\n! nbrf format database %s opened: %d sequences\n", name,
	    db->sequences);

    /* deal with adding it to the cache of DATABASE entries. */
    add_database(db);

    return db;
}

/**
 * Opens a, FASTA format, sequence file. This builds an internal database
 * cache entry (which must be released by the user when no longer required)
 * that contains its name, type and the file handles that contain
 * the sequence and sequence-index data. This will return a
 * NULL if any of the files (data, index) cannot be opened.

 * @see open_embl_database, open_nbrf_database
 *
 * @param name is the base-name of the file containing a FASTA format
 *     sequence database. The name will have '.fasta' and '.index' appended
 *     and it expects to find both files available for reading.
 *
 * @returns a DATABASE handle if the file is successfully opened and NULL
 *     otherwise. 
 */
DATABASE *
open_fasta_database(char *name)
{
    char buf[256];
    DATABASE *db;

    db = (DATABASE *) calloc(1, sizeof(DATABASE));
    db->database = (char *) strdup(name);
    db->datafile = openfile_in_seqpath(name, "fasta", "r", buf);
    db->indexfile = openfile_in_seqpath(name, "index", "r", buf);
    db->type = FASTA;

    if (!db->datafile || !db->indexfile) {
	free(db);
	return NULL;
    }
    make_fasta_index(db);

    fprintf(stderr, "\n! fasta format database %s opened: %d sequences\n", name,
	    db->sequences);

    add_database(db);

    return db;
}

/**
 * Caches a database record. Currently stores the entries in a linear, fixed
 * length, table. If the capacity is exceeded the an error is generated and
 * the calling application exits.
 *
 * @param db is the input parameter specifying the database record to cache.
 */
static void
add_database(DATABASE *db)
{
    if (database_count < MAX_DB_COUNT)
	databases[database_count++] = db;
    else {
	fprintf(stderr, "\nERROR: too many databases! \n");
	exit(1);
    }
}

/**
 * Load data from a EMBL index file into a in-memory hash table. The hash
 * table is constructed using dynamicaly allocated memory (which must be
 * released by the user when no longer required. The hash table comprises
 * a sequence entry-name and file offset pair.
 *
 * @param db is the database cache entry that should be updated when the 
 *     index data is read in. This also supplies the FILE stream that should
 *     be used to read the index data (pre computed by the calling routine).
 */
void
make_embl_index(DATABASE *db)
{
    int n;
    unsigned long offset;
    HASH_LIST *item;
    char name[256];

    db->index = (HASH_LIST **) calloc(PRIME_NUM, sizeof(HASH_LIST *));

    while (fscanf(db->indexfile, "%s %lu\n", name, &offset) == 2) {
	n = hash_name(name);
	item = (HASH_LIST *) calloc(1, sizeof(HASH_LIST));
	item->name = (char *) strdup(name);
	item->offset = offset;
	db->sequences++;

	if (!db->index[n])
	    db->index[n] = item;
	else {
	    item->next = db->index[n];
	    db->index[n] = item;
	}
    }

    rewind(db->indexfile);
}

/**
 * Load data from a FASTA index file into a in-memory hash table. The hash
 * table is constructed using dynamicaly allocated memory (which must be
 * released by the user when no longer required. The hash table comprises
 * a sequence entry-name and file offset pair.
 *
 * @param db is the database cache entry that should be updated when the 
 *     index data is read in. This also supplies the FILE stream that should
 *     be used to read the index data (pre computed by the calling routine).
 */
void
make_fasta_index(DATABASE *db)
{
    int n;
    unsigned long offset;
    HASH_LIST *item;
    char name[256];

    db->index = (HASH_LIST **) calloc(PRIME_NUM, sizeof(HASH_LIST *));

    while (fscanf(db->indexfile, "%s %lu\n", name, &offset) == 2) {
	n = hash_name(name);
	item = (HASH_LIST *) calloc(1, sizeof(HASH_LIST));
	item->name = (char *) strdup(name);
	item->offset = offset;
	db->sequences++;

	if (!db->index[n])
	    db->index[n] = item;
	else {
	    item->next = db->index[n];
	    db->index[n] = item;
	}
    }

    rewind(db->indexfile);
}

/**
 * Load data from a NBRF index file into a in-memory hash table. The hash
 * table is constructed using dynamicaly allocated memory (which must be
 * released by the user when no longer required. The hash table comprises
 * a sequence entry-name and file offset pair.
 *
 * @param db is the database cache entry that should be updated when the 
 *     index data is read in. This also supplies the FILE stream that should
 *     be used to read the index data (pre computed by the calling routine).
 */
void
make_nbrf_index(DATABASE *db)
{
    int n;
    unsigned long offset1, offset2;
    HASH_LIST *item;
    char name[256];

    db->index = (HASH_LIST **) calloc(PRIME_NUM, sizeof(HASH_LIST *));

    while (fscanf(db->indexfile, "%s %lu %lu\n", name, &offset1, &offset2) == 3) {
	n = hash_name(name);
	item = (HASH_LIST *) calloc(1, sizeof(HASH_LIST));
	item->name = (char *) strdup(name);
	item->offset = offset1;
	item->text_offset = offset2;
	db->sequences++;

	if (!db->index[n])
	    db->index[n] = item;
	else {
	    item->next = db->index[n];
	    db->index[n] = item;
	}
    }

    rewind(db->indexfile);
}

/**
 * Opens a database and reads a named sequence from that source.
 *
 * @param name is a string inthe form "seqname:dbname". This specifies
 *     both the name of the database containing the sequence and the 
 *     name of the desired sequence.
 *
 * @returns the SEQUENCE object which corresponds to the supplied 
 *    sequence name. If the routine cannot determine the type of the
 *    database handle that the user supplied then this routine
 *    will return NULL.
 */
SEQUENCE *
read_seq(char *name)
{
    char *db, *sname;
    DATABASE *d;

    sname = (char *) strchr(name, ':') + 1;
    *(sname - 1) = 0;
    db = name;

    d = open_database(db);
    return read_sequence(sname, d);

}

/**
 * Fetch the named sequence from the supplied database handle. This
 * is a wrapper routine that makes the decision of which specalised
 * sequence reader should be used based on the type fiedl of the 
 * DATABASE record supplied to it.
 *
 * @param name is the name of the sequence required.
 * @param database is the DATABASE handle that contains the required
 *     sequence data.
 *
 * @returns the SEQUENCE object which corresponds to the supplied name.
 *    If the routine cannot determine the type of the database handle
 *    that the user supplied then this routine will return NULL.
 */
SEQUENCE *
read_sequence(char *name, DATABASE *database)
{
    unsigned long text_offset;
    unsigned long offset = get_offset(name, database, &text_offset);

    if (database->type == EMBL)
	return read_embl_sq(name, database, offset);
    else if (database->type == NBRF)
	return read_nbrf_sq(name, database, offset, text_offset);
    else if (database->type == FASTA)
	return read_fasta_sq(name, database, offset);
    else
	return NULL;
}

/**
 * Examines the in-memory hash table associated with the supplied DATABASE
 * handle for the sequence with the supplied name. The offset into the
 * sequence data file is returned through the text_offset parameter.
 *
 * @param name is the name of teh desired sequence.
 * @param database is the DATABASE handle for the database containing the
 *     desired sequence.
 * @param name text_offset is a (unsigned long) return parameter that 
 *     will be bound to the offset within the data file where the start
 *     of the sequence is located. 
 *
 * @returns the offset of the sequence within the data file associated with
 *     the DATABASE or zero if the sequence name cannot be found.
 */
unsigned long
get_offset(char *name, DATABASE *database, unsigned long *text_offset)
{
    int n;
    HASH_LIST *h;

    n = hash_name(name);
    h = database->index[n];

    *text_offset = 0;

    while (h) {
	if (!strcmp(name, h->name)) {
	    *text_offset = h->text_offset;
	    return h->offset;
	} else
	    h = h->next;
    }
    return 0;
}


/**
 * Computes an integer hash value between 0 and PRIME_NUM-1 for the
 * supplied character string (case-insensitive). <b>FILE PRIVATE</b>
 *
 * @param string is the null terminated string for which the has value
 *     should be computed.
 *
 * @returns an integer has value in the range 0..PRIME_MUN-1.
 */
static int
hash_name(char *string)
{

    int n = 0;

    while (*string) {
	n = (64 * n + tolower((int) *string)) % PRIME_NUM;
	string++;
    }

    return n;
}


/**
 * Extracts a named sequence from the supplied DATABASE using the supplied
 * location. A sanity check is made to ensure that the offset is valid by
 * confirming the sequence name. The description and sequence data are
 * then loaded from the "DE" and "SQ" fields. The sequence data is read
 * until either a '/' terminator is found or the neccessary number of 
 * bases have been read (as specified by the "SQ    Sequence <len>"
 * string). All base codes are mapped to lower case. <b>NB</b>
 * this routine dynamically allocates storage for the sequence, name
 * description (always null in this case) and the SEQUENCE structure.
 * This storage will need to be released by the user when it is no
 * longer required.
 *
 * @param name is the name of the sequence and should correspond to the
 *     "ID" field of the EMBL record format.
 * @param database is the DATABASE handle for the database which should
 *     be searched for the desired sequence.
 * @param offset is the offset within the data file that should be at the
 *     start of the desired sequence.
 *
 * @returns a SEQUENCE object (which has storage dynamically allocated
 *     by this routine) that contains the actual sequence data.
 *     If the supplied name and the name extracted from the data file
 *     don't match then an error is written to standard error and the
 *     calling application is terminated.
 */
SEQUENCE *
read_embl_sq(char *name, DATABASE *database, unsigned long offset)
{
    char sname[256], line[512];
    int c;
    SEQUENCE *seq = (SEQUENCE *) calloc(1, sizeof(SEQUENCE));
    int len, n;

    fseek(database->datafile, offset, 0);

    find_next(database->datafile, "ID", line);

    *sname = 0;
    sscanf(line, "ID  %s", sname);

    if (strcmp(sname, name)) {
	fprintf(stderr,
		"\n! ERROR on read_embl_sq: pointer offset for %s wrong! (gets %s instead)\n",
		name, sname);
	exit(1);
    }
    seq->name = (char *) strdup(name);
    seq->database = database;

    find_next(database->datafile, "DE", line);

    seq->desc = (char *) strdup(&line[5]);

    find_next(database->datafile, "SQ", line);

    sscanf(line, "SQ   Sequence %d", &len);

    seq->len = len;
    seq->s = (char *) calloc(len + 1, sizeof(char));
    n = 0;

    while ((c = getc(database->datafile)) != '/' && c != EOF) {
	if (isalpha(c) && !isspace(c) && n < len)
	    seq->s[n++] = (char) tolower(c);
    }
    return seq;
}

/**
 * Extracts a named sequence from the supplied DATABASE using the supplied
 * location. A sanity check is made to ensure that the offset is valid by
 * confirming the sequence name. A two pass approach is adopted to determine
 * the length of the sequence data and then read it in. Sequence is read
 * until the next '>' character or End Of File is reached. <b>NB</b>
 * this routine dynamically allocates storage for the sequence, name
 * description (always null in this case) and the SEQUENCE structure.
 * This storage will need to be released by the user when it is no
 * longer required.
 * 
 * All base codes are mapped to lower case.
 *
 * @param name is the name of the sequence and should correspond to the
 *     "><name>" field of the FASTA record format.
 * @param database is the DATABASE handle for the database which should
 *     be searched for the desired sequence.
 * @param offset is the offset within the data file that should be at the
 *     start of the desired sequence.
 *
 * @returns a SEQUENCE object (which has storage dynamically allocated
 *     by this routine) that contains the actual sequence data.
 *     If the supplied name and the name extracted from the data file
 *     don't match then an error is written to standard error and the
 *     calling application is terminated.
 */
SEQUENCE *
read_fasta_sq(char *name, DATABASE *database, unsigned long offset)
{
    char sname[256], line[512];
    int c;
    SEQUENCE *seq = (SEQUENCE *) calloc(1, sizeof(SEQUENCE));
    int len, n;
    unsigned long current;

    fseek(database->datafile, offset, 0);

    tts_read_line(database->datafile, line);
    *sname = 0;
    sscanf(line, ">%s", sname);

    if (strcmp(sname, name)) {
	fprintf(stderr,
		"\n! ERROR on read_fasta_sq: pointer offset for %s wrong! (gets %s instead)\n",
		name, sname);
	exit(1);
    }
    seq->name = (char *) strdup(name);
    seq->database = database;

    seq->desc = (char *) strdup("");

    current = ftell(database->datafile);

    /*
     * Do one scan to determine the length of the sequence. Terminator for
     * the multi sequence files is the next '>' character. May also terminate
     * at the end of the file.
     */
       
    len = 0;
    while ((c = getc(database->datafile)) != '>' && c != EOF)
	len += (!isspace(c) && c != '\n');
    seq->len = len;
    seq->s = (char *) calloc(len + 1, sizeof(char));
    n = 0;

    /* Reset to beginning of the sequence data */
    fseek(database->datafile, current, 0);

    while ((c = getc(database->datafile)) != '>' && c != EOF)
	if (!isspace(c) && c != '\n' && n < len)
	    seq->s[n++] = (char) tolower(c);

    seq->len = n;

    return seq;
}

/**
 * Extracts a named sequence from the supplied DATABASE using the supplied
 * location. A sanity check is made to ensure that the offset is valid by
 * confirming the sequence name. A two pass approach is adopted to determine
 * the length of the sequence data and then read it in. Sequence is read
 * until the next '>' character or End Of File is reached. <b>NB</b>
 * this routine dynamically allocates storage for the sequence, name
 * description (always null in this case) and the SEQUENCE structure.
 * This storage will need to be released by the user when it is no
 * longer required.
 *
 * <b>NB</b>
 * this routine dynamically allocates storage for the sequence, name
 * description (always null in this case) and the SEQUENCE structure.
 * This storage will need to be released by the user when it is no
 * longer required.
 *
 * @param name is the name of the sequence and should correspond to the
 *     "ID" field of the EMBL record format.
 * @param database is the DATABASE handle for the database which should
 *     be searched for the desired sequence.
 * @param offset is the offset within the data file that should be at the
 *     start of the desired sequence.
 *
 * @returns a SEQUENCE object (which has storage dynamically allocated
 *     by this routine) that contains the actual sequence data.
 *     If the supplied name and the name extracted from the data file
 *     don't match then an error is written to standard error and the
 *     calling application is terminated.
 */
SEQUENCE *
read_nbrf_sq(char *name, DATABASE *database,
             unsigned long offset, unsigned long text_offset)
{
    char sname[256], line[512];
    int c;
    SEQUENCE *seq = (SEQUENCE *) calloc(1, sizeof(SEQUENCE));
    unsigned long current;
    int len;

    fseek(database->datafile, offset, 0);

    /* Extract the sequence name from the sequence data file */
    *sname = 0;
    tts_read_line(database->datafile, line);
    sscanf(line, ">>>>%s", sname);

    if (strcmp(sname, name)) {
	fprintf(stderr,
		"\n! ERROR on read_nbrf_sq: pointer offset for %s wrong! (gets %s instead)\n",
		name, sname);
	exit(1);
    }
    seq->name = (char *) strdup(name);
    seq->database = database;

    tts_read_line(database->datafile, line);
    seq->desc = (char *) strdup(line);

    current = ftell(database->datafile);
    len = 0;

    /*
     * Do one scan to determine the length of the sequence. Terminator for
     * the multi sequence files is the next '>' character. May also terminate
     * at the end of the file.
     */
    while ((c = getc(database->datafile)) != '>' && c != EOF) {
	if (isalpha(c) && !isspace(c))
	    len++;
    }

    seq->len = len;
    seq->s = (char *) calloc(len + 1, sizeof(char));

    /* Return to the original position */
    fseek(database->datafile, current, 0);
    len = 0;

    /* Read the neccessary number of sequence bases into the allocated store */
    while ((c = getc(database->datafile)) != '>' && c != EOF) {
	if (isalpha(c) && !isspace(c))
	    seq->s[len++] = (char) tolower(c);
    }

    return seq;
}

/**
 * Searches through a database index file returning consecutive matching
 * sequences according to the supplied spec string. This string is of the
 * form "db:seqname[*]". If an entry is found in the database (db) index
 * then read_sequence() will be called to load the actual sequence data.
 *
 * @param spec is a string of the form "db:seqname[*]" which indicates the
 * name of the database file containing the sequence data and the sequence
 * name (with an optional, pseudo-wildcard character) that is required.
 *
 * @returns the first (and subsequent matching) SEQUENCE objects. If no
 * matching sequence is found the NULL is returned.
 */
SEQUENCE *
next_seq_from_database_spec(char *spec)
{
    char wild[256], line[256], name[256];
    DATABASE *db = is_sequence_spec(spec, wild);
    int len = strlen(wild);

    while (tts_read_line(db->indexfile, line) != EOF) {
	sscanf(line, "%s", name);
	if (!strncmp(name, wild, len))
	    return read_sequence(name, db);
    }

    return NULL;
}

/**
 * This is a statefull routine that will open a sequence database (if
 * required). It is statefull in the sense that if the routine is
 * called multiple times with the same spec then it will not reopen
 * the database. It has a memory of one spec.
 *
 * @param spec is the specification for a database. This is of the form
 *     "db:sname[*]".
 * @param wild is a string return parameter which is bound to the second
 *     part of the spec string (following the ':' character) and terminating
 *     at the first '*' character.
 *
 * @returns a DATABASE handle to the database name specified in the first
 *     part of the spec string.
 */
DATABASE *
is_sequence_spec(char *spec, char *wild)
{
    static char *old_spec = NULL;
    static char *old_wild;
    static DATABASE *old_db;
    char buf[256], *sname, *db, *s;

    /*
     * Either the first call (old_spec == NULL) or the supplied spec does
     * not match the current state of this routine.
     */
    if (!old_spec || strcmp(spec, old_spec)) {
	old_spec = (char *) strdup(spec);
	strcpy(buf, spec);

	/* Look for the position of the ':' character and split the string */
	sname = (char *) strchr(buf, ':');
	if (!sname) {
	    sname = (char *) strdup("");
	} else {
	    sname++;
	    *(sname - 1) = 0;
	}

	/*
         * At this point we have:
         *
         * buf -  first part of the spec string (upto ':')
         * sname - second part of the spec string (or null string)
         */

	/*
         * If the second part of the string contains a '*' - wildcard?
         * terminate the string at that point
         */
	if ((s = strchr(sname, '*')) != 0)
	    *s = 0;

	db = buf;
	old_wild = (char *) strdup(sname);
	old_db = open_database(db);
    }

    /*
     * old_wild and old_db only get changed if the spec is different or
     * its the first call of the routine. 
     */

    strcpy(wild, old_wild);
    return old_db;
}

/**
 * Using a specification string of the form "db:sname[*]" this routine
 * performs a rewind of both the data and index files associated with the
 * database named in the first part of the specification string. If
 * the spec fails to result in is_sequence_spec() yielding a valid
 * DATABASE handle then an attempt is made to interpret the spec as a
 * file name and perform a rewind on that.
 *
 * @param spec a database and sequence name specification string of the 
 *     form "db:sname[*]".
 */
void
rewind_spec(char *spec)
{
    char buf[256];
    DATABASE *db = is_sequence_spec(spec, buf);
    FILE *fp;

    if (db) {
	rewind(db->indexfile);
	rewind(db->datafile);
    } else if ((fp = which_file_of_sequences(spec)) != 0)
	rewind(fp);
}


/**
 * Returns repeated (on multiple invocations of this routine) sequence
 * objects from the spec_or_file data source. This can either be a
 * string of the form "db:seqname[*]" or a file name with a '@' symbol
 * prepended.
 *
 * @param spec_or_file is a string containing a specification for either
 *     a file or a sequence specifier
 *
 * @returns a sequence object or NULL if an error occurs while trying
 *     to open the database/file and/or reading a sequence with the
 *     appropriate name.
 */
SEQUENCE *
next_seq(char *spec_or_file)
{
    char wild[256];
    DATABASE *db = is_sequence_spec(spec_or_file, wild);
    FILE *fp;

    if (db)
	return next_seq_from_database_spec(spec_or_file);
    else {
	if (*spec_or_file == '@'
	    && (fp = which_file_of_sequences(spec_or_file)))
	    return next_seq_from_list_file(fp);
    }
    return 0;
}

/**
 * Returns a FILE stream to the supplied file name. This routine also
 * manages a private cache of currently open files. The file name is
 * used as the key to a linear table of FILE streams.
 *
 * @param filename is the name of the required file with a '@' character
 * prepended.
 *
 * @returns the FILE stream object associated with the supplied filename.
 *     If an excessive number of files are open then this routine will
 *     force the termination of the calling program.
 */
FILE *
which_file_of_sequences(char *filename)
{
    static char *filenames[100];
    static FILE *fp[100];
    static int files;
    int n;

    if (*filename != '@')
	return NULL;
    else
	filename++;

    for (n = 0; n < files; n++) {
	if (!strcmp(filenames[n], filename)) {
	    if (fp[n])
		return fp[n];
	    else
		return fp[n] = seq_openfile(filename, "r");
	}
    }

    if (files == 100) {
	fprintf(stderr, "ERROR: too many files of sequences!\n");
	exit(1);
    }

    filenames[files] = (char *) strdup(filename);
    fp[files] = seq_openfile(filename, "r");
    return fp[files++];
}
