/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

#include <stdio.h>
#include <stdlib.h>

#include <io_lib/scf.h>
#include "sequtils.h"

static const char *cvsid = "$Id: seq_display.c 18816 2006-08-03 16:39:42Z rmd $";

/**
 * Generates a Postscript (Level 3) rendition of a SCF trace. This is
 * optionaly annotated by the addition of grey lines that indicate where
 * base transitions have occured and, if available, add the base codes
 * on the X axis.
 *
 * @param fp is where the generated postscript code should be written
 * @param scf is the Scf data source containing the trace data.
 * @param trace_start is the integer location within the trace where the
 *     graphing should start.
 * @param trace_end is the integer location within the trace where the
 *     graphing should end.
 * @param base_count is the number of bases in the base array.
 * @param base is a char array containing the base codes that should be
 *     marked on the graph. If this is null then no bases will be added
 *     to the graph.
 * @param base_start is the start index in the trace_coord array for the
 *     first base that should be marked on the trace graph.
 * @param base_end is the end index in the trace_coord array that indicates
 *     the last base that should be marked on the trace graph.
 * @param trace_coord is an array of trace coordinates where each base
 *     transition should be marked. This is done by adding a grey line
 *     to the trace at each point. If the parameter base is also available
 *     the the actual base code is added along the axis.
 *     If this is null then no base information will be displayed.
 * @param title is the title string for the trace output. This will be 
 *     displayed on each page out output (together with the page number)
 *     in 14pt Helvetica-Bold.
 */
void
ps_print_alignment(FILE *fp, Scf *scf, int trace_start, int trace_end,
		   int base_count, char *base, int base_start, int base_end,
		   int *trace_coord, char *title) {

    int pixels_per_line = 500;
    int lines_per_page = 4;
    int pages;
    int page;
    int line;
    int base_incr = 10;
    int page_height = 800;
    int height = page_height / lines_per_page - 20;
    int h = height - 50;
    int trace_position = trace_start;
    Samples1 *s1;
    Samples2 *s2;
    int i, x;
    float scale = (scf->header.sample_size == 2) ? (h / 65536.0) : (h / 256.0);
    int *trace_coord2 = NULL;
    float p;

    /* 
     * Build an array (of the same length as the trace) that contains a value
     * of one wherever trace_coord indicates that there is a base change.
     * The parameter trace_coord lists the trace offset where each base
     * should be marked.
     */
    if (trace_coord) {
	trace_coord2 = (int *) calloc(trace_end + 1, sizeof(int));
	for (i = base_start; i <= base_end; i++) {
	    trace_coord2[trace_coord[i]] = 1;
	}
    }
    p = (trace_end - trace_start) / (1.0 + pixels_per_line * lines_per_page);
    if ((float) ((int) p) < p)
	pages = p + 1;
    else
	pages = p;

    ps_intro(fp, pages);

    for (page = 1; page <= pages; page++) {
	fprintf(fp, "/Helvetica-Bold findfont 14 scalefont setfont 0 setgray\n");
	fprintf(fp, "%%%%Page: %d %d\n%%%%BeginPageSetup\n%%%%EndPageSetup\nnewpath\n", page, page);
	fprintf(fp, "%5d %5d translate\n", 70, page_height - height);
	fprintf(fp, "%5d %5d m (%s     page %d/%d) sh\n", 0, 150, (title ? title : ""), page, pages);
	fprintf(fp, "/Helvetica-Bold findfont 10 scalefont setfont\n");

	for (line = 0; line < lines_per_page; line++) {
	    fprintf(fp, "\n%% line %d\n", line + 1);
	    if (scf->header.sample_size == 2) {		/* 16-bit data */
		s2 = &scf->samples.samples2[trace_position];
		fprintf(fp, "gs adenine 0 %g m\n", s2->sample_A * scale);
		for (i = 0;
		     i < pixels_per_line && trace_position + i <= trace_end;
		     i++, s2++)
		    fprintf(fp, "%d %g l\n", i, s2->sample_A * scale);
		fprintf(fp, "s gr\n");

		s2 = &scf->samples.samples2[trace_position];
		fprintf(fp, "gs cytosine 0 %g m\n", s2->sample_C * scale);
		for (i = 0;
		     i < pixels_per_line && trace_position + i <= trace_end;
		     i++, s2++)
		    fprintf(fp, "%d %g l\n", i, s2->sample_C * scale);
		fprintf(fp, "s gr\n");

		s2 = &scf->samples.samples2[trace_position];
		fprintf(fp, "gs guanine 0 %g m\n", s2->sample_G * scale);
		for (i = 0;
		     i < pixels_per_line && trace_position + i <= trace_end;
		     i++, s2++)
		    fprintf(fp, "%d %g l\n", i, s2->sample_G * scale);
		fprintf(fp, "s gr\n");

		s2 = &scf->samples.samples2[trace_position];
		fprintf(fp, "gs thyamine 0 %g m\n", s2->sample_T * scale);
		for (i = 0;
		     i < pixels_per_line && trace_position + i <= trace_end;
		     i++, s2++)
		    fprintf(fp, "%d %g l\n", i, s2->sample_T * scale);
		fprintf(fp, "s gr\n");
	    } else {		/* 8 bit data */
		s1 = &scf->samples.samples1[trace_position];
		fprintf(fp, "gs adenine 0 %g m\n", s1->sample_A * scale);
		for (i = 0;
		     i < pixels_per_line && trace_position + i <= trace_end;
		     i++, s1++)
		    fprintf(fp, "%d %g l\n", i, s1->sample_A * scale);
		fprintf(fp, "s gr\n");

		s1 = &scf->samples.samples1[trace_position];
		fprintf(fp, "gs cytosine 0 %g m\n", s1->sample_C * scale);
		for (i = 0;
		     i < pixels_per_line && trace_position + i <= trace_end;
		     i++, s1++)
		    fprintf(fp, "%d %g l\n", i, s1->sample_C * scale);
		fprintf(fp, "s gr\n");

		s1 = &scf->samples.samples1[trace_position];
		fprintf(fp, "gs guanine 0 %g m\n", s1->sample_G * scale);
		for (i = 0;
		     i < pixels_per_line && trace_position + i <= trace_end;
		     i++, s1++)
		    fprintf(fp, "%d %g l\n", i, s1->sample_G * scale);
		fprintf(fp, "s gr\n");

		s1 = &scf->samples.samples1[trace_position];
		fprintf(fp, "gs thyamine 0 %g m\n", s1->sample_T * scale);
		for (i = 0;
		     i < pixels_per_line && trace_position + i <= trace_end;
		     i++, s1++)
		    fprintf(fp, "%d %g l\n", i, s1->sample_T * scale);
		fprintf(fp, "s gr\n");
	    }

	    if (trace_coord) {
		/* s1 = &scf->samples.samples1[trace_position]; */
		fprintf(fp, "gs unknown\n");
		for (i = 0; i < pixels_per_line && trace_position + i <= trace_end; i++ /*,s1++ */ )
		    if (trace_coord2[i + trace_position])
			fprintf(fp, "%5d 0 m %5d %d l s\n", i, i, h);
	    }
	    if (base && trace_coord) {
		for (i = base_start; i <= base_end; i++) {
		    if (trace_coord[i] && trace_coord[i + 1] >= trace_position && trace_coord[i + 1] < trace_position + pixels_per_line) {
			if (trace_coord[i + 1] > trace_coord[i])
			    x = trace_coord[i + 1] - trace_coord[i];
			else
			    x = 10;
			fprintf(fp, "%d (%c) wid  %d add %d m (%c) sh\n", x, base[i], trace_coord[i] - trace_position, h, base[i]);
			if (!(i % base_incr)) {
			    x = trace_coord[i] - trace_position;
			    fprintf(fp, "gs black %d %d m (%d) sh gr\n", x, h - 12, i);
			    fprintf(fp, "gs black %d %d m (%d) sh gr\n", x, -12, trace_coord[i]);
			}
		    }
		}
	    }
	    fprintf(fp, "0 %d translate\n", -height);
	    trace_position += pixels_per_line;
	}
	fprintf(fp, "\nshowpage\n");
    }
    fprintf(fp, "%%%%Trailer\n%%%%EOF\n");

    if (trace_coord2)
	free(trace_coord2);
}

/**
 * Generates a Postscript (Level 3) rendition of a SCF trace. This is
 * optionaly annotated by the addition of grey lines that indicate where
 * base transitions have occured and, if available, add the base codes
 * on the X axis. This version differs from ps_print_alignment() in that
 * it can be called multiple times to generate multiple traces within
 * one document. These can, optionally, be on different pages.
 *
 * @param fp is where the generated postscript code should be written
 * @param scf is the Scf data source containing the trace data.
 * @param trace_start is the integer location within the trace where the
 *     graphing should start.
 * @param trace_end is the integer location within the trace where the
 *     graphing should end.
 * @param base_count is the number of bases in the base array.
 * @param base is a char array containing the base codes that should be
 *     marked on the graph. If this is null then no bases will be added
 *     to the graph.
 * @param base_start is the start index in the trace_coord array for the
 *     first base that should be marked on the trace graph.
 * @param base_end is the end index in the trace_coord array that indicates
 *     the last base that should be marked on the trace graph.
 * @param centre is a parameter, the effect of which, is not clear.
 * @param trace_coord is an array of trace coordinates where each base
 *     transition should be marked. This is done by adding a grey line
 *     to the trace at each point. If the parameter base is also available
 *     the the actual base code is added along the axis.
 *     If this is null then no base information will be displayed.
 * @param title is the title string for the trace output. This will be 
 *     displayed on each page out output (together with the page number)
 *     in 14pt Helvetica-Bold.
 * @param linetitle is not used.
 * @param newpage is a flag that, when true, indicates that the routine
 *     should generate a 'showpage' and hence force a new output page if
 *     this routine has previously output any content.
 */
void
ps_print_multiple_alignment(FILE *fp, Scf *scf, int trace_start, int trace_end,
			    int base_count, char *base, int base_start,
			    int base_end, int centre, int *trace_coord,
			    char *title, char *linetitle, int *newpage) {
    int pixels_per_line = 400;
    int lines_per_page = 8;
    static int pages = 0;
    int page = 0;
    static int line = 0;
    int base_incr = 10;
    int page_height = 800;
    int height = page_height / lines_per_page - 20;
    int h = height - 50;
    int trace_position = trace_start;
    Samples1 *s1;
    Samples2 *s2;
    int i, x;
    float scale = (scf->header.sample_size == 2) ? (h / 65536.0) : (h / 256.0);
    int *trace_coord2 = NULL;
    float p;

    if (trace_coord) {
	trace_coord2 = (int *) calloc(trace_end + 1, sizeof(int));
	for (i = base_start; i <= base_end; i++) {
	    trace_coord2[trace_coord[i] + trace_start] = 1;
	}
    }
    p = (trace_end - trace_start) / (1.0 + pixels_per_line * lines_per_page);

    if (pages == 0)
	ps_intro(fp, 0);

    if (*newpage == 1) {
	*newpage = 0;

	if (page > 0)
	    fprintf(fp, "\nshowpage\n");

	page++;
	pages++;

	fprintf(fp, "/Helvetica-Bold findfont 14 scalefont setfont 0 setgray\n");
	fprintf(fp, "%%%%Page: %d %d\n%%%%BeginPageSetup\n%%%%EndPageSetup\nnewpath\n", page, page);
	fprintf(fp, "%5d %5d translate\n", 70, page_height - height);
	fprintf(fp, "%5d %5d m (%s     page %d) sh\n", 0, 150, (title ? title : ""), page);
	fprintf(fp, "/Helvetica-Bold findfont 10 scalefont setfont\n");

	line = 0;
    }
    trace_position = centre - 0.5 * pixels_per_line;

    fprintf(fp, "\n%% line %d\n", line + 1);
    if (scf->header.sample_size == 2) {		/* 16-bit data */
	s2 = &scf->samples.samples2[trace_position];
	fprintf(fp, "gs adenine 0 %g m\n", s2->sample_A * scale);
	for (i = 0;
	     i < pixels_per_line && trace_position + i <= trace_end;
	     i++, s2++)
	    fprintf(fp, "%d %g l\n", i, s2->sample_A * scale);
	fprintf(fp, "s gr\n");

	s2 = &scf->samples.samples2[trace_position];
	fprintf(fp, "gs cytosine 0 %g m\n", s2->sample_C * scale);
	for (i = 0;
	     i < pixels_per_line && trace_position + i <= trace_end;
	     i++, s2++)
	    fprintf(fp, "%d %g l\n", i, s2->sample_C * scale);
	fprintf(fp, "s gr\n");

	s2 = &scf->samples.samples2[trace_position];
	fprintf(fp, "gs guanine 0 %g m\n", s2->sample_G * scale);
	for (i = 0;
	     i < pixels_per_line && trace_position + i <= trace_end;
	     i++, s2++)
	    fprintf(fp, "%d %g l\n", i, s2->sample_G * scale);
	fprintf(fp, "s gr\n");

	s2 = &scf->samples.samples2[trace_position];
	fprintf(fp, "gs thyamine 0 %g m\n", s2->sample_T * scale);
	for (i = 0;
	     i < pixels_per_line && trace_position + i <= trace_end;
	     i++, s2++)
	    fprintf(fp, "%d %g l\n", i, s2->sample_T * scale);
	fprintf(fp, "s gr\n");
    } else {			/* 8 bit data */
	s1 = &scf->samples.samples1[trace_position];
	fprintf(fp, "gs adenine 0 %g m\n", s1->sample_A * scale);
	for (i = 0;
	     i < pixels_per_line && trace_position + i <= trace_end;
	     i++, s1++)
	    fprintf(fp, "%d %g l\n", i, s1->sample_A * scale);
	fprintf(fp, "s gr\n");

	s1 = &scf->samples.samples1[trace_position];
	fprintf(fp, "gs cytosine 0 %g m\n", s1->sample_C * scale);
	for (i = 0;
	     i < pixels_per_line && trace_position + i <= trace_end;
	     i++, s1++)
	    fprintf(fp, "%d %g l\n", i, s1->sample_C * scale);
	fprintf(fp, "s gr\n");

	s1 = &scf->samples.samples1[trace_position];
	fprintf(fp, "gs guanine 0 %g m\n", s1->sample_G * scale);
	for (i = 0;
	     i < pixels_per_line && trace_position + i <= trace_end;
	     i++, s1++)
	    fprintf(fp, "%d %g l\n", i, s1->sample_G * scale);
	fprintf(fp, "s gr\n");

	s1 = &scf->samples.samples1[trace_position];
	fprintf(fp, "gs thyamine 0 %g m\n", s1->sample_T * scale);
	for (i = 0;
	     i < pixels_per_line && trace_position + i <= trace_end;
	     i++, s1++)
	    fprintf(fp, "%d %g l\n", i, s1->sample_T * scale);
	fprintf(fp, "s gr\n");
    }

    if (trace_coord) {
	/* s1 = &scf->samples.samples1[trace_position]; */
	fprintf(fp, "gs unknown\n");
	for (i = 0; i < pixels_per_line && trace_position + i <= trace_end; i++ /*,s1++ */ )
	    if (trace_coord2[i + trace_position])
		fprintf(fp, "%5d 0 m %5d %d l s\n", i, i, h);
    }
    if (base && trace_coord) {
	for (i = base_start; i <= base_end; i++) {
	    if (trace_coord[i] && trace_coord[i + 1] + trace_start >= trace_position && trace_coord[i + 1] + trace_start < trace_position + pixels_per_line) {
		if (trace_coord[i + 1] > trace_coord[i])
		    x = trace_coord[i + 1] - trace_coord[i];
		else
		    x = 10;
		fprintf(fp, "%d (%c) wid  %d add %d m (%c) sh\n", x, base[i], trace_coord[i] + trace_start - trace_position, h, base[i]);
		if (!(i % base_incr)) {
		    x = trace_coord[i] + trace_start - trace_position;
		    fprintf(fp, "gs black %d %d m (%d) sh gr\n", x, h - 12, i);
		    fprintf(fp, "gs black %d %d m (%d) sh gr\n", x, -12, trace_coord[i] + trace_start);
		}
	    }
	}
    }
    fprintf(fp, "0 %d translate\n", -height);

    if (trace_coord2)
	free(trace_coord2);
}

/**
 * Writes a Postscript (level 3) header and a series of postscript
 * macros for use by the graph plotting routines.
 *
 * @param fp is the file stream to write the postscript data into.
 * @param pages is the total number of pages within the document.
 */
void
ps_intro(FILE *fp, int pages)
{
    if (pages)
	fprintf(fp, "%%!PS-Adobe-3.0\n%%%%Pages: %d\n%%%%EndComments\n%%%%BeginProlog\n", pages);
    else
	fprintf(fp, "%%!PS-Adobe-3.0\n%%%%Pages: (atend)\n%%%%EndComments\n%%%%BeginProlog\n");

    fprintf(fp, "/adenine  { 0 1 0 setrgbcolor } def\n");
    fprintf(fp, "/cytosine { 0 0 1 setrgbcolor } def\n");
    fprintf(fp, "/guanine  { 0 0 0 setrgbcolor } def\n");
    fprintf(fp, "/thyamine { 1 0 0 setrgbcolor } def\n");
    fprintf(fp, "/unknown  { 0.5 0.5 0.5 setrgbcolor } def\n");
    fprintf(fp, "/black    { 0 0 0 setrgbcolor } def\n");
    fprintf(fp, "/gs { gsave } def\n");
    fprintf(fp, "/gr { grestore } def\n");
    fprintf(fp, "/s  { stroke } def\n");
    fprintf(fp, "/sh { show } def\n");
    fprintf(fp, "/m { moveto } def\n");
    fprintf(fp, "/l { lineto } def\n");
    fprintf(fp, "/wid { stringwidth pop /w exch def w sub 2 div } def\n");
    fprintf(fp, "%%%%EndProlog\n%%%%BeginSetup\n");
    fprintf(fp, "%%%%EndSetup\n");
}
