/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

/**
 * @file
 * This file defines routines associated with file IO related to sequence
 * data.
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "sequtils.h"
#include "searchpath.h"
#include <readline.h>
#include <cl.h>

static const char *cvsid = "$Id: seq_io.c 35448 2010-07-01 15:08:05Z jkb $";

/**
 * Read a sequence from file. The format of the file is (with a new
 * line separating items): name, description and finally the sequence
 * data. The terminator for the sequence is EOF. White space and non-alpha
 * characters within the sequence data are ignored. Dynamic allocation
 * is used for storage of the sequence structure, name, description
 * and sequence strings. The maximum sequence length that can be read by
 * this function is MAXSEQLEN.
 *
 * @param filename is the file name for the file containing the sequence
 * data.
 *
 * @returns a pointer to a sequence 'object'
 */
SEQUENCE *
get_seq_from_file(char *filename)
{
    FILE *fp;
    SEQUENCE *seq;
    char name[256];
    char desc[256];
    char buf[MAXSEQLEN + 1];
    char *s, *t;
    int len;

    if ((fp = seq_openfile(filename, "r")) != 0) {
	seq = (SEQUENCE *) calloc(1, sizeof(SEQUENCE));
	tts_read_line(fp, desc);
	sscanf(desc, "%s", name);
	seq->name = (char *) strdup(name);
	tts_read_line(fp, desc);
	seq->desc = (char *) strdup(desc);
	len = fread(buf, sizeof(char), MAXSEQLEN, fp);
	buf[len] = 0;
	s = buf;
	t = buf;
	len = 0;
	while (*s) {
	    if (isalpha((int) *s)) {
		*t++ = tolower((int) *s);
		len++;
	    }
	    s++;
	}
	*t = 0;
	seq->len = len;
	seq->s = (char *) strdup(buf);
	fclose(fp);

#ifdef DEBUG
        fprintf(stderr, "read sequence %s len %d from file %s\n",
                name, len, filename );
#endif

	return seq;
    }
    return NULL;
}

/**
 * Attempts to open filename, using mode to determine function
 * (read/write/append), and terminates the program on failure.
 * This is a copy, private to sequtils, derived from the libcl 
 * version. This has been done to remove the dependency on the 
 * command line processing library which may not always be used
 * in future.
 *
 * @param filename is the name of the file to be opened
 * @param mode is one of "read", "write" or "append" indicating the desired
 * I/O mode.
 * 
 * @returns a FILE pointer to the opened file (on failure the calling
 * application will be terminated and this will not return).
 */
FILE *
seq_openfile(char *filename, char *mode)
{
    FILE *fp;

    if ((fp = fopen(seq_make_legal(filename), mode)) == NULL) {
	fprintf(stderr, "\n ERROR: file %s will not open for mode %s\n\n", filename, mode);
	exit(1);
    } else
	return fp;
}

/*
 * use compiler switch -DVAX to use the form of make_legal which converts 
 * '.' to '_'
 */

#ifdef VAX

/**
 * Translates file names of the form file.in.out into file_in.out and hence
 * into a format suitable for use on VAX systems. <b>NB</b> this is a 
 * destructive operator that modifies the original instance of the string.
 *
 * @param filename is the string containing the file name to be modified.
 *
 * @returns the modified version of the string with all dots (except the
 *     last one) changed to underscores.
 */
static char *
seq_make_legal(char *filename)

{
    char *s = filename, *t;
    int n = 0;

    while (*s) {
	if (*s == '.') {
	    n++;
	    t = s;
	}
	s++;
    }

    if (n > 1) {
	t--;
	while (t != filename) {
	    if (*t == '.')
		*t = '_';
	    t--;
	}
    }
    return filename;
}


#else

/**
 * Null version of the seq_make_legal file name translation for use on systems
 * which can cope with multiple dots in file names.
 *
 * @see make_legal_vax
 *
 * @param filename is the string containing the file name to be modified.
 *
 * @returns the unmodified version of filename.
 */
static char *
seq_make_legal(char *filename)
{
    return filename;
}

#endif

/**
 * Read a FASTA sequence in from a file. Can be used for multiple reads 
 * from a file (positional state is held by the file stream). An internal,
 * static, buffer is used to hold a temporary copy of the sequence data as
 * it is read in. The SEQUENCE structure and the name, desc and s fields are
 * all dynamically allocated and should be freed by the user when no longer
 * required.
 *
 * @param fp is the file stream that should be read from.
 *
 * @returns a SEQUENCE object containing the first sequence that can be
 *     found by starting a read at the current position in the stream.
 */
SEQUENCE *
get_fasta_sq(FILE * fp)
{
    int c;
    SEQUENCE *seq = NULL;
    char name[256], desc[256];
    char *s;
    int i;
    static char *buffer = NULL;
    static int buflen = 0;

    /*
     * Allocates a default 100K buffer. This gets enlarged if required
     * and will exist for the lifetime of the calling program. This buffer
     * is used for all reads and hence gets overwritten on each invocation.
     */

    if (buffer == NULL) {
	buflen = 100000;
	buffer = (char *) calloc(buflen, sizeof(char));
    }

    /*
     * Terminating condition for each sequence record is the location of a
     * '>' character (or end of file). 
     */

    while ((c = getc(fp)) != '>' && c != EOF);

    if (c == EOF) /* Cannot get here ? */
	return NULL;
    else {

	/*
         * Construct a SEQUENCE structure to associate the sequence data
         * with.
         */
	seq = (SEQUENCE *) calloc(1, sizeof(SEQUENCE));

	/*
         * Read the name of the sequence from the file. Clip at 254 bytes.
         */
	i = 0;
	while (!isspace(c = getc(fp)))
	    if (i < 254)
		name[i++] = c;

	/*
         * Null terminate and make a fresh copy on the heap that the seq
         * structure can reference.
         */
	name[i] = 0;
	seq->name = strdup(name);

	/*
         * Skip any spaces until we hit the next newline.
         */
	while (c != '\n' && c != EOF && isspace(c = getc(fp)));


        /*
         * Read the sequence description fromthe file. Clip at 254 bytes.
         */
	i = 0;
	while (c != '\n' && c != EOF) {
	    if (i < 254)
		desc[i++] = c;
	    c = getc(fp);
	}

	/*
         * Null terminate and make a fresh copy on the heap that the seq
         * structure can reference.
         */
	desc[i] = 0;
	seq->desc = strdup(desc);

        /*
         * Read the actual sequence data into the buffer (resize it if 
         * necessary). Only accepts alpha characters and the '-' character
         * so all white space is ignored.
         */

	seq->len = 0;
	while ((c = getc(fp)) != '>' && c != EOF) {
	    if (isalpha(c) || (c == '-')) {

                /*
                 * We have reached capacity. Resize the buffer.
                 */
		if (seq->len + 1 >= buflen) {
		    buflen += 100000;
		    buffer = (char *) realloc(buffer, buflen * sizeof(char));
		}
		buffer[seq->len++] = c;
	    }
	}

	/*
         * Null terminate and make a fresh copy on the heap that the seq
         * structure can reference.
         */
	buffer[seq->len] = 0;
	s = seq->s = (char *) strdup(buffer);

	/* Make the buffer ready for next use. */
	buffer[0] = 0;

	/* Push back the last character '>' for next read */
	ungetc(c, fp);
    }
    return seq;
}

/**
 * Search for the location of a text string within the supplied file
 * stream. The check is done on a line-by-line basis and any non-printable
 * characters found in the file are replaced by spaces. The actual matching
 * line is stored in the buffer passed to the routine as the line parameter.
 *
 * @param fp is the file stream containing the data to be searched.
 * @param text is the target string that should be searched for.
 * @param line is a return parameter (string buffer) that should be large
 *     enough to hold a full line from the input stream.
 *
 * @returns the offset into the file where the string occurs or zero if
 *     the string is not found. <b>NB.</b> this routine does not reset the
 *     file pointer associated with the stream.
 */
unsigned long
find_next(FILE *fp, char *text, char *line)
{
    while (tts_read_line(fp, line) != EOF) {
	clean_line(line);
	if (!strncmp(line, text, strlen(text)))
	    return ftell(fp) - strlen(line);
    }
    return 0;
}

