/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

/**
 * @file
 * This file defines routines associated with manipulating sequence 'objects'
 * withing the sequence utilities package.
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "sequtils.h"

static const char *cvsid = "$Id: seq_ops.c 18816 2006-08-03 16:39:42Z rmd $";

static char *complement_table = NULL;

/**
 * Return a character representing the complement-base of the supplied
 * parameter.  The mapping is as follows: a->t, c->g, g->c, t|u->a, [->],
 * ]->[, -->-, all others->n. There is a single shot initalisation
 * of a static lookup table to represent this mapping. In addition the
 * case of the supplied parameter is preserved, Uppercase->Uppercase and
 * vice-versa.
 *
 * @param c is the character representing a base. Mapped values include:
 * {a,c,g,t,u,[,],-} all other inputs get a default mapping of 'n'.
 *
 * @returns the character representation of the bilogical compliment of the
 * supplied base. 
 */
char
complement_base(char c)
{

    if (!complement_table) {
	int x;
	complement_table = (char *) calloc(256, sizeof(char)) + 127;

	for (x = -127; x < 128; x++) {
	    if (x == 'a')
		complement_table[x] = 't';
	    else if (x == 'c')
		complement_table[x] = 'g';
	    else if (x == 'g')
		complement_table[x] = 'c';
	    else if (x == 't' || x == 'u')
		complement_table[x] = 'a';
	    else if (x == '[')
		complement_table[x] = ']';
	    else if (x == ']')
		complement_table[x] = '[';
	    else if (x == '*')
		complement_table[x] = '*';
	    else if (x == '-')
		complement_table[x] = '-';
	    else
		complement_table[x] = 'n';
	}
    }

    if (isupper((int) c))
	return toupper((int) complement_table[tolower((int) c)]);
    else
	return complement_table[tolower((int) c)];
}

/**
 * Convert a string containing a string representation of a base sequence
 * into its biological complement. See complement_base for the mapping. This
 * also reverses the direction of the string.
 *
 * @see complement_base
 *
 * @param seq is the input sequence. <b>NB.</b> this is destructively modified.
 *
 * @returns a pointer to the original string storage but with the contents
 * modified to have complement characters.
 */
char *
complement_seq(char *seq)
{
    char *s = seq;
    char c, *t;

    while (*s) {
	*s = complement_base(*s);

	s++;
    }

    t = seq;
    s--;

    while (t < s) {
	c = *s;
	*s = *t;
	*t = c;
	t++;
	s--;
    }

    return seq;
}

/**
 * Creates a SEQUENCE containing the subsequence start to stop inclusive
 *
 * If start > stop then returns the reverse complement
 *
 * If start and stop are out of range they are truncated
 *
 */
SEQUENCE *
subseq(SEQUENCE * seq, int start, int stop)

{
    SEQUENCE *subs = seqdup(seq);
    char *s;
    int i;
    int reverse;

    if (start > stop) {
	reverse = 1;
	i = start;
	start = stop;
	stop = i;
    } else
	reverse = 0;

    if (start < 0)
	start = 0;
    if (start > seq->len - 1)
	return NULL;
    if (stop > seq->len - 1)
	stop = seq->len - 1;

    subs->len = stop - start + 1;
    s = (char *) calloc(subs->len + 1, sizeof(char));

    for (i = 0; i < subs->len; i++)
	s[i] = seq->s[start + i];

    if (reverse)
	complement_seq(s);

    free(subs->s);
    subs->s = s;

    return subs;
}

/**
 * Duplicate a sequence object. Allocates new storage and makes copies of
 * its name and description provided they were non-null in the source
 * sequence.
 *
 * @param seq is the source sequence to be duplicated.
 *
 * @returns a pointer to a SEQUENCE struture with a deep copy of name, desc
 * and s elements. If the supplied sequence is NULL then this routine will
 * return NULL also.
 */
SEQUENCE *
seqdup(SEQUENCE * seq)
{
    SEQUENCE *dup;

    if (seq) {
	dup = (SEQUENCE *) calloc(1, sizeof(SEQUENCE));
	*dup = *seq;

	if (seq->name)
	    dup->name = (char *) strdup(seq->name);
	else
	    dup->name = NULL;

	if (seq->desc)
	    dup->desc = (char *) strdup(seq->desc);
	else
	    dup->desc = NULL;

	if (seq->s)
	    dup->s = (char *) strdup(seq->s);
	else
	    dup->s = NULL;

	return dup;
    }
    return NULL;
}


/**
 * Construct a SEQUENCE 'object'. Builds the data structure, assigns
 * values to each of the components and performs some internal house
 * functions keeping.
 *
 * @see STRUCTURE
 *
 * @param name is the name of this sequence 'object'
 * @param desc is the description string
 * @param s is the sequence string data.
 *
 * @retuns a pointer to the new SEQUENCE object.
 */
SEQUENCE *
into_sequence(char *name, char *desc, char *s)
{
    SEQUENCE *seq = (SEQUENCE *) calloc(1, sizeof(SEQUENCE));
    seq->name = name;
    seq->desc = desc;
    seq->s = s;
    seq->len = strlen(s);
    return seq;
}

/**
 * In-place pseudo-random shuffling of a sequence. See shuffle_s for a
 * description of the algorithim used.
 * 
 * @see shuffle_s
 *
 * @param seq is the original sequence who's sequence string should be
 * randomly permuted.
 * @param in_place is a boolean flag indicating whether the sequence should
 * duplicated before permutation or whether this should be a destructive
 * operator. False=duplicate before permutation.
 * @param seed is the random number generator seed.
 *
 * @returns the original sequence (or a copy) with the string, representing
 * its base sequence, randomly permuted.
 */
SEQUENCE *
shuffle_seq(SEQUENCE * seq, int in_place, int *seed)
{
    SEQUENCE *shuffled;

    if (!in_place)
	shuffled = seqdup(seq);
    else
	shuffled = seq;

    shuffle_s(shuffled->s, seed);

    return shuffled;
}


/**
 * PACKAGE PRIVATE- data structure to be used only in the kv_cmp and
 * shuffle_s routines.
 * @see kv_cmp, shuffle_s
 */
typedef struct
{
  float key;
  int value;
}
key_value;

/**
 * PACKAGE PRIVATE- sort comparison function used by qsort to compare two
 * key_value strucures based on their respective key's.
 *
 * @param aa is the pointer to the first key-value pair
 * @param bb is the pointer to the second key-value pair
 * 
 * @returns the result of fcmp being run on the key's of the two pairs.
 */
static int
kv_cmp(const void *aa, const void *bb)
{
    const key_value *a = aa;
    const key_value *b = bb;
    return fcmp((float*)&(a->key), (float*)&(b->key));
}

/**
 * Compares two float values passed by reference. <b>NB.</b> there
 * may be an issue with 'identical' floats not returning 0 from this
 * routine due to the classic comparison of floats error. This probably
 * has little effect on sorting.
 *
 * @param a is the first reference (pointer to an float value)
 * @param b is the second reference (pointer to an float value)
 *
 * @returns 1 if a>b, -1 if a<b and 0 if a=b
 */
static int fcmp(float *a, float *b) {
  float x = *a - *b;
  if ( x > 0.0 )
    return 1;
  else if ( x < 0.0 )
    return -1;
  else
    return 0;
}

extern float rand3(int*); /* rand3.c in ../utils */

/**
 * In-place pseudo-random shuffling of a string. Makes key-value pairs of 
 * the form: <random, s[n]> then qsorts the set of these values based on
 * their random key. The supplied string is then overwritten with the value
 * components of the sorted set of these key-value pairs (in ascending order
 * of key).
 * 
 * @see kv_cmp, key_value 
 *
 * @param s is the original string to randomly permute.
 * @param seed is the random number generator seed.
 *
 * @returns the original string randomly permuted by the process described 
 * above.
 */
static char *
shuffle_s(char *s, int *seed)
{

    key_value *tmp;
    int n;
    int len = strlen(s);

    tmp = (key_value *) calloc(len, sizeof(key_value));

    for (n = 0; n < len; n++) {
	tmp[n].key = rand3(seed);
	tmp[n].value = s[n];
    }

    qsort(tmp, len, sizeof(key_value), kv_cmp);

    for (n = 0; n < len; n++)
	s[n] = tmp[n].value;

    free(tmp);

    return s;
}
