/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

/* $Id: seq_util.c 18811 2006-08-03 16:33:28Z rmd $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "sequtils.h"

static const char *cvsid = "$Id: seq_util.c 18811 2006-08-03 16:33:28Z rmd $";

/**
 * Recovers the accession number from a sequence object/file. This routine
 * determines what type of sequence lies behind the SEQUENCE object and
 * forwards the call to one of embl_seq_comment or nbrf_seq_comment.
 *
 * @see embl_seq_comment, nbrf_seq_comment
 *
 * @param seq is the SEQUENCE object to query for the accession number.
 *
 * @returns a string containing the accession number.
 */
char *
acc_nos(SEQUENCE *seq)
{

    if (seq->database->type == EMBL)
	return embl_seq_comment(seq, "AC");
    else
	return nbrf_seq_comment(seq, "ACCESSION");
}

/**
 * This routine returns the contents of any comment field associated with
 * the supplied SEQUENCE object. This actually acts as a call forwarder
 * after determining the underlying sequence type. The call is forwarded
 * to one of embl_seq_comment or nbrf_seq_comment.
 *
 * @see embl_seq_comment, nbrf_seq_comment
 *
 * @param seq is the SEQUENCE object to query for the comment data.
 * @param key is the key to use when searching the file for matching
 *     fields. In EMBL files this will be "DE".
 *
 * @returns a string containing the comment field from the sequence.
 */
char *
seq_comment(SEQUENCE *seq, char *key)
{

    if (seq->database->type == EMBL)
	return embl_seq_comment(seq, key);
    else
	return nbrf_seq_comment(seq, key);
}

/**
 * Extract and return the comment field associated with the supplied
 * SEQUENCE object. The key value should (I suspect) always be set to
 * "DE" for this type of file.

 * <b>NB</b> the returned string is a pointer to a static array
 * declared within the routine. This may cause unexpected behaviour
 * if multiple references are held to this pointer and the routine
 * is called multiple times.
 *
 * @param seq is the SEQUENCE object to query for the comment data.
 * @param key is the key to use when searching the file for matching
 *     fields. In EMBL files this will be "DE".
 *
 * @returns a string containing the comment field  from the sequence.
 */
char *
embl_seq_comment(SEQUENCE *seq, char *key)
{

    char line[256];
    unsigned long text_offset;
    unsigned long offset = get_offset(seq->name, seq->database, &text_offset);
    unsigned long end_pos;
    static char buf[MAX_BUF];

    *buf = 0;
    fseek(seq->database->datafile, offset, 0);

    /* '//' marks the end of a embl file, find its offset. */
    end_pos = find_next(seq->database->datafile, "//", line);
    fseek(seq->database->datafile, offset, 0);

    while (1) {
	find_next(seq->database->datafile, key, line);
	if (end_pos < ftell(seq->database->datafile))
	    break;
	else {
	    if (strlen(buf) + strlen(line) + 2 < MAX_BUF)
		strcat(buf, clean_line(line) + strlen(key));
	    else
		break;
	}
    }

    return buf;
}

/**
 * Find and extract the comment field from a NBRF format sequence file.
 * It is not clear what the value of key should be for this type of file
 * but looking at examples suggests that ">DL" will pick up the header 
 * data for a sequence.
 *
 * @param seq is the SEQUENCE object to query for the comment data.
 * @param key is the key to use when searching the file for matching
 *     fields. 
 *
 * @returns a string containing the comment field  from the sequence.
 */
char *
nbrf_seq_comment(SEQUENCE * seq, char *key)
{

    char line[256];
    unsigned long text_offset = 0;
    unsigned long end_pos;
    static char buf[MAX_BUF];

    *buf = 0;
    if (seq->database->textfile) {
	fseek(seq->database->textfile, text_offset + 4, 0);
	end_pos = find_next(seq->database->textfile, ">>>>", line);
	fseek(seq->database->textfile, text_offset, 0);

	while (1) {
	    find_next(seq->database->textfile, key, line);
	    if (end_pos < ftell(seq->database->textfile))
		break;
	    else {
		if (strlen(buf) + strlen(line) + 2 < MAX_BUF)
		    strcat(buf, clean_line(line) + strlen(key));
		else
		    break;
	    }
	}
    }
    return buf;
}

/**
 * 'Parse' a file in EMBL format and generate (in the same location as the
 * .dat file) an .index file. This contains a record of the byte offset where
 * each EMBL 'ID' string resides together with a copy of the identifier
 * string without the leading 'ID' string.
 * 
 * The state machine used in this routine simply identifies a 'ID' string
 * when it starts at column zero.
 * See http://www.psc.edu/general/software/packages/seq-intro/emblfile.html
 * for the significance of this.
 *
 * @param name is the name of the file, in EMBL format, to be indexed. This
 * should be the basename (i.e. not have the '.dat' extension.
 *
 */
void
compile_embl_index(char *name)
{
    char buf[256];
    FILE *datafile, *indexfile;
    char sname[256];
    char c;
    unsigned long n, offset;
    int state = -1;

    datafile = openfile_in_seqpath(name, "dat", "r", buf);
    indexfile = openfile_in_seqpath(name, "index", "w", buf);

    fprintf(stderr, "compiling embl format index for database %s ...\n", name);

    state = -1;
    offset = 0;
    n = 0;
    while ((c = fgetc(datafile)) != EOF) {
	if (c == '\n') {
	    state = 1;
	} else if ((state == 1 || state == -1) && c == 'I') {
	    state = 2;
	} else if (state == 2 && c == 'D') {
	    offset = ftell(datafile) - 2;
	    fscanf(datafile, "%s", sname);
	    n++;
	    fprintf(indexfile, "%-10s %lu\n", sname, offset);
	    printf("%-10s %lu\n", sname, offset);
	    state = 0;
	} else
	    state = 0;
    }
    fclose(indexfile);
    fclose(datafile);

    fprintf(stderr, "... done. %lu sequences indexed\n", n);
}

/**
 * 'Parse' a file in FASTA format and generate (in the same location as the
 * FASTA file) an .index file. This contains a record of the byte offset where
 * each description string lay in the FASTA file together with a copy of the
 * description string without the leading '>' character.
 * 
 * The state machine used in this routine simply identifies a '>' character
 * when it is in column zero. See http://www.ncbi.nlm.nih.gov/BLAST/fasta.html
 * for the significance of this.
 *
 * @param name is the name of the file, in FASTA format, to be indexed. This
 * should be the basename (i.e. not have the '.fasta' extension.
 *
 */
void
compile_fasta_index(char *name)
{
    char buf[256];
    FILE *datafile, *indexfile;
    char sname[256];
    char c;
    unsigned long n, offset;
    int state = -1;

    datafile = openfile_in_seqpath(name, "fasta", "r", buf);
    indexfile = openfile_in_seqpath(name, "index", "w", buf);

    fprintf(stderr, "compiling fasta format index for database %s ...\n", name);

    state = -1;
    offset = 0;
    n = 0;
    while ((c = fgetc(datafile)) != EOF) {
	if (c == '\n') {
	    state = 1;
	} else if ((state == 1 || state == -1) && c == '>') {
	    state = 2;
	} else
	    state = 0;

	if (state == 2) {
	    offset = ftell(datafile) - 1;
	    fscanf(datafile, "%s", sname);
	    n++;
	    fprintf(indexfile, "%-10s %lu\n", sname, offset);
	    printf("%-10s %lu\n", sname, offset);
	    state = 0;
	}
    }
    fclose(indexfile);
    fclose(datafile);

    fprintf(stderr, "... done. %lu sequences indexed\n", n);
}

/**
 * 'Parse' files in NBRF format and generate (in the same location as the pair
 * of NBRF files) an .index file. This contains a record of the byte offset
 * where each '>>>>' string lay in the .seq and .ref files together with
 * the name that immediately follows the '>>>>' pattern. An additional check
 * is made to ensure that bothe the .seq and .ref files have the names in
 * the same sequential order.
 *
 * NB. NBRF stands for 'National Biomedical Research Foundation'
 * 
 * @param name is the name of the file, in NBRF format, to be indexed. This
 * should be the basename (i.e. not have the '.seq' or '.ref' extension.
 *
 */
void
compile_nbrf_index(char *name)
{
    char buf[256];
    FILE *datafile, *textfile, *indexfile;
    char name1[256], name2[256];
    long c, d;
    unsigned long n;

    datafile = openfile_in_seqpath(name, "seq", "r", buf);
    textfile = openfile_in_seqpath(name, "ref", "r", buf);
    indexfile = openfile_in_seqpath(name, "index", "w", buf);

    fprintf(stderr, "compiling nbrf format index for database %s ...\n", name);

    n = 0;
    while (1) {
	c = seekto(datafile, ">>>>");
	d = seekto(textfile, ">>>>");

	if (c == EOF || d == EOF)
	    break;

	fscanf(datafile, ">>>>%s", name1);
	fscanf(textfile, ">>>>%s", name2);

	if (strcmp(name1, name2)) {
	    fprintf(stderr, "\n! ERROR when indexing: different names %s %s\n",
		    name1, name2);
	    exit(1);
	}

	fprintf(indexfile, "%-10s %10lu %10lu\n", name1, c, d);
	n++;
    }
    fclose(indexfile);
    fclose(datafile);
    fclose(textfile);

    fprintf(stderr, "... done. %lu sequences indexed\n", n);
}

/**
 * Seek through a file until the next occurence of the text string occurs.
 * as a side effect the file pointer will be located (via fseek) to point
 * at the start of the target string.
 *
 * @param fp is the file stream to be searched.
 * @param text is the string to be matched.
 *
 * @returns EOF if this occurs before a successful match occurs. 
 * Otherwise the byte offset of the start of the target string
 * (relative to the start of the FILE) is returned.
 */
unsigned long
seekto(FILE * fp, char *text)
{
    unsigned long p;
    char c, *s;

    while (1) {
	c = getc(fp);
	if (c == EOF)
	    return EOF;
	else if (c == *text) {
	    p = ftell(fp);
	    s = text;
	    while (*s) {
		if (c != *s)
		    break;
		else if (c == EOF)
		    return EOF;
		else {
		    s++;
		    c = getc(fp);
		}
	    }
	    if (!*s) {
		p--;
		fseek(fp, p, 0);
		return p;
	    } else
		fseek(fp, p + 1, 0);
	}
    }
}

/**
 * Deallocate all storage associated with the supplied structure. Assumes
 * that the sequence name, string and description are all allocated storage
 * on the heap (i.e. malloc'd)
 *
 * @param seq is the sequence object to be destroyed.
 */
void
free_seq(SEQUENCE * seq)
{
    if (seq) {
	if (seq->name)
	    free(seq->name);
	if (seq->s)
	    free(seq->s);
	if (seq->desc)
	    free(seq->desc);
	free(seq);
    }
}

/**
 * Convert all characters in the supplied string into lower case.
 *
 * @param s is the input string.
 *
 * @returns a pointer to the original string (which has been modified
 * to have only lower case characters.
 */
char *
downcase(char *s)
{

    char *t = s;
    while (*s) {
	*s = tolower((int) *s);
	s++;
    }
    return t;
}


/**
 * Convert all characters in the supplied string into upper case.
 *
 * @param s is the input string.
 *
 * @returns a pointer to the original string (which has been modified
 * to have only upper case characters.
 */
char *
upcase(char *s)
{

    char *t = s;
    while (*s) {
	*s = toupper((int) *s);
	s++;
    }
    return t;
}


/**
 * Replace any non-printable characters in a string with a space character.
 *
 * @param s is the input string.
 *
 * @returns a pointer to the original string (which has been modified
 * to have no unprintable characters in it.
 */
char *
clean_line(char *s)
{

    char *t = s;

    while (*s) {
	if (!isprint((int) *s))
	    *s = ' ';
	s++;
    }
    return t;
}
