/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

#include <stdio.h>

#include <io_lib/scf.h>

#ifndef _SEQUTIL_H_
#define _SEQUTIL_H_

typedef enum {
    EMBL, NBRF, FASTA, SINGLE
} WHAT_KIND_OF_DATABASE;

typedef struct hash_list {
    char *name;
    unsigned long offset, text_offset;
    struct hash_list *next;
} HASH_LIST;

typedef struct {
    char *database;
    WHAT_KIND_OF_DATABASE type;
    int sequences;
    unsigned long length;
    FILE *datafile;
    FILE *textfile;
    FILE *indexfile;
    HASH_LIST **index;
} DATABASE;

typedef struct {
    char *name;
    DATABASE *database;
    char *desc;
    int len;
    char *s;
} SEQUENCE;

/* seq_db.c */

#define PRIME_NUM 10007

/*
 * Used in the database cache array. Represents the maximum number of databases
 * that can be open simultaniously.
 */
#define MAX_DB_COUNT 100

DATABASE *open_database(char *name);
DATABASE *which_database(char *database_name);
DATABASE *open_embl_database(char *name);
DATABASE *open_nbrf_database(char *name);
DATABASE *open_fasta_database(char *name);

void make_embl_index(DATABASE * db);
void make_fasta_index(DATABASE * db);
void make_nbrf_index(DATABASE * db);

SEQUENCE *read_seq(char *name);
SEQUENCE *read_sequence(char *name, DATABASE * database);
unsigned long get_offset(char *name, DATABASE * database, unsigned long *text_offset);
SEQUENCE *read_embl_sq(char *name, DATABASE * database, unsigned long offset);
SEQUENCE *read_fasta_sq(char *name, DATABASE * database, unsigned long offset);
SEQUENCE *read_nbrf_sq(char *name, DATABASE * database, unsigned long offset, unsigned long text_offset);
SEQUENCE *next_seq_from_database_spec(char *spec);
DATABASE *is_sequence_spec(char *spec, char *wild);
void rewind_spec(char *spec);
SEQUENCE *next_seq(char *spec_or_file);
FILE *which_file_of_sequences(char *filename);


static void add_database(DATABASE * db);
static int hash_name(char *string);

/* seq_io.c */

#define MAXSEQLEN 100000

FILE *openfile_in_seqpath_arg(char *format, char *ext,
			   char *mode, int argc, char **argv, char *fullname);
FILE *openfile_in_seqpath(char *basename, char *ext,
			  char *mode, char *fullname);
SEQUENCE *next_seq_from_list_file(FILE * list_file);
SEQUENCE *get_seq_from_file(char *filename);
SEQUENCE *get_fasta_sq(FILE * fp);
unsigned long find_next(FILE * fp, char *text, char *line);
FILE * seq_openfile(char *filename, char *mode);
static char * seq_make_legal(char *filename);

/* seq_ops.c */

char complement_base(char c);
char *complement_seq(char *seq);
SEQUENCE *subseq(SEQUENCE * seq, int start, int stop);
SEQUENCE *seqdup(SEQUENCE * seq);
SEQUENCE *into_sequence(char *name, char *desc, char *s);
SEQUENCE *shuffle_seq(SEQUENCE * seq, int in_place, int *seed);

static int kv_cmp(const void *aa, const void *bb);
static int fcmp(float *a, float *b);
static char *shuffle_s(char *s, int *seed);

/* seq_util.c */

#define MAX_BUF 10000

char *acc_nos(SEQUENCE * seq);

char *seq_comment(SEQUENCE * seq, char *key);
char *embl_seq_comment(SEQUENCE * seq, char *key);
char *nbrf_seq_comment(SEQUENCE * seq, char *key);
unsigned long seekto(FILE * fp, char *text);

void compile_embl_index(char *name);
void compile_fasta_index(char *name);
void compile_nbrf_index(char *name);

void free_seq(SEQUENCE * seq);
char *downcase(char *s);
char *upcase(char *s);
char *clean_line(char *s);

/* seq_display.c */
void  ps_print_alignment (FILE *fp, Scf *scf, int trace_start, int trace_end, int base_count, char *base, int base_start, int base_end, int *trace_coord, char *title);

void  ps_print_multiple_alignment (FILE *fp, Scf *scf, int trace_start, int trace_end, int base_count, char *base, int base_start, int base_end, int centre, int *trace_coord, char *title, char *linetitle, int *newpage);

void  ps_intro (FILE *fp, int pages);

#endif
