/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include "sequtils.h"
#include "sw.h"

static const char *cvsid = "$Id: sw.c 10638 2003-04-15 09:15:12Z dgm $";

int sim_mat[256][256];
static int mat_init_ok = 0;

/**
 * Unpack two SEQUENCE structures and perform a Smith Waterman alignement
 * of one sequence against another. The score and the start and finish
 * coordinates of the optimal local aligmnent are returned by the routine
 * and output parameters coord1 and coord2.
 *
 * @param seq1 is a SEQUENCE object representing one of the sequences to
 *     perform alignment with.
 * @param seq2 is a SEQUENCE object representing the second sequence to
 *     perform alignment against.
 * @param match is the integer score that should be attributed to an exact
 *     sequence element match.
 * @param mismatch is the integer score that should be attributed to a
 *     sequence element mis-match.
 * @param gap is the integer penalty that should be ascribed to the creation
 *     of a gap in either sequence.
 * @param neutral is the score attributed to a pair of 'N's being matched
 * @param offset
 * @param window
 * @param coord1
 * @param coord2
 *
 * @returns 
 */
int 
smith_waterman_score(SEQUENCE * seq1, SEQUENCE * seq2, int match, int mismatch, int gap, int neutral, int offset, int window, int *coord1, int *coord2)
{

    /*
     * Unpack the two SEQUENCE structures to obtain their raw charcter string
     * sequence data and sequence lengths.
     */
    int len1 = seq1->len;
    int len2 = seq2->len;
    char *s1 = seq1->s;
    char *s2 = seq2->s;

    alignStruct *as = smith_waterman_alignment2(s1, s2, len1, len2, match,
                           mismatch, gap, neutral, offset, window);

    int score = smith_waterman_score2(s1, s2, len1, len2, match,
                           mismatch, gap, neutral, offset, window,
                           coord1, coord2);

    int ok = ((score == as->score) &&
             (*coord1 == as->end1) &&
             (*coord2 == as->end2));

    printf("checking: score: %d %d coord1: %d %d coord2: %d %d  ok:%d\n",
        score, as->score, *coord1, as->end1, *coord2, as->end2, ok);

    return score;
}


/**
 * Returns the smith-waterman score for the sequences s1 and s2, in a
 * band of half-width=window around diagonal=offset (=seq1-seq2) matches
 * score +match, mismatches +mismatch, gaps +gap (no gap start penalty)
 *
 * @param s1 is the null terminated string representing the first sequence
 *     to be aligned. <b>NB.</b> this is the sequence that forms the horizontal
 *     axis of the alignment matrix.
 * @param s2 is the null terminated string representing the second sequence
 *     to be aligned. <b>NB.</b> this is the sequence that forms the vertical
 *     axis of the alignment matrix.
 * @param len1 is the length if the sequence supplied in the parameter s1.
 * @param len2 is the length if the sequence supplied in the parameter s2.
 * @param match the the score that should be attributed to two exactly
 *     matching symbols drawn from s1 and s2.
 * @param mismatch is the mismatch penalty that should be associated with
 *     two mismatched symbols  from s1 and s2.
 * @param gap is the gap penalty associated with introducing a gap into
 *     either of the sequences.
 * @param neutral is the score attributed to a pair of matching 'n' 
 *     symbols in the two sequences.
 * @param offset is the offset into either s1 or s2 (if the offset is
 *     negative) that the banded alignment should be centered on.
 * @param window is the window size for banded alignment. The actual width
 *     of the region of the SW matrix explored by this routine is 2*window+1.
 *     A negative value for the window size signals that full alignment
 *     (without banding) should be carried out.
 * @param end_coord1 is the location in sequence s1 where the maximum Smith
 *     Waterman score is found.
 * @param end_coord2 is the location in sequence s2 where the maximum Smith
 *     Waterman score is found.
 *
 * @returns the maximum score found during the alignment. The actual terminal
 *     coordinates for the alignment can be found by inspecting end_coord1
 *     and end_coord2
 */
int 
smith_waterman_score2(char *s1, char *s2, int len1, int len2,
                      int match, int mismatch, int gap, int neutral,
                      int offset, int window, int *end_coord1,
                      int *end_coord2)
{
    int *A, *B, *a, *b, *c;
    int i1, i2;
    int max_score = 0;
    int c1 = 0, c2 = 0;
    int m;
    int d, u, l;
    int start1, stop1;
    int start2, stop2;
    int x2;
    char *lower;
    char pad_char = '*';	/* NOTE fix this! */

    /* Single shot substitution matrix initalisation */
    if (!mat_init_ok)
	mat_init(match, mismatch, gap, neutral, pad_char);

    /*
     * len1 is corrected to be the the length os the supplied string if
     * its less than the value supplied by the user.
     */
    if (len1 > (m = strlen(s1)))
	len1 = m;

    /*
     * len2 is corrected to be the the length os the supplied string if
     * its less than the value supplied by the user.
     */
    if (len2 > (m = strlen(s2)))
	len2 = m;

    /*
     * Allocate space for two rows in the alignment matrix. The +1 is used
     * to allow an array to index -1 for the boundary condition. The calloc
     * also ensures that the elements of the array are zero.
     */
    A = (int *) calloc(len1 + 1, sizeof(int));
    B = (int *) calloc(len1 + 1, sizeof(int));

    /* Map everything in the first sequence string to lower case */
    lower = (char *) downcase(strdup(s1));

    /* Make reference arrays that are valid in the range [-1..len1] */
    a = A + 1;
    b = B + 1;

    /*
     * Banded alignment. Flag for this happening is a value for window >= 0.
     *
     * Default is to cover the full length of sequence 2.
     *
     * start2/stop2 are the row index bounds for exploring the SW matrix.
     *     The value taken on by each of these is bounded by zero and the
     *     length of the second sequence.
     *
     * NB. For +ve offset this describes a band which starts at a point 
     * that is offset into sequence1. A -ve offset applies the offset to
     * sequence2.
     *
     * e.g For offset = +7 and window=3 the following region of the SW
     * matrix is evaluated. Note the clipping of the row length as it
     * approaches the end of sequence 1. Note also indexed from zero.
     *
     *                  seq1
     *
     *     A B C D E F G H I J K L M N O
     *     B       X X X X X X X        
     *     C         X X X X X X X      
     *     D           X X X X X X X    
     *  s  E             X X X X X X X  
     *  e  F               X X X X X X X
     *  q  G                 X X X X X X
     *  2  H                   X X X X X
     *     I                     X X X X
     *     J                       X X X
     *     K                         X X
     *     L                           X
     *     M                            
     *     N                            
     *     O                            
     */
    if (window < 0) {
	start2 = 0;
	stop2 = len2 - 1;
    } else {
	start2 = -offset - window;
	if (start2 < 0)
	    start2 = 0;
	if (start2 >= len2)
	    start2 = len2 - 1;

	stop2 = len1 - offset + window;
	if (stop2 < 0)
	    stop2 = 0;
	if (stop2 >= len2)
	    stop2 = len2 - 1;
    }

    /*
     * Perform the walk down each of the rows of the SW matrix. a and b get
     * swapped over on each iteration.
     * a - Previous row of Matrix. (a[-1] = 0, by use of calloc A)
     * b - Current row of the Matrix. (b[-1] = 0, by use of calloc B)
     */
    for (i2 = start2; i2 <= stop2; i2++) {
	c = a;
	a = b;
	b = c;
	if (window < 0) {
	    start1 = 0;
	    stop1 = len1 - 1;
	} else {
	    start1 = i2 + offset - window;
	    if (start1 < 0)
		start1 = 0;
	    if (start1 >= len1)
		start1 = len1 - 1;
	    stop1 = i2 + offset + window;
	    if (stop1 < 0)
		stop1 = 0;
	    if (stop1 >= len1)
		stop1 = len1 - 1;
	}
	x2 = tolower((int) s2[i2]);

	for (i1 = start1; i1 <= stop1; i1++) {

            /*
             * Compute the components of the max() term of the SW recurrence.
             *     d - is the diagonal (consume 1 symbol of each sequence).
             *     u - up
             *     l - left
             */

	    d = a[i1 - 1] + sim_mat[(int) lower[i1]][x2];
	    u = a[i1] + gap;
	    l = b[i1 - 1] + gap;

            /*
             * MAX of the possibilities is chosen.
             */
	    if (d > u) {
		if (d > l) {
                     /* d holds the MAX value */
		    b[i1] = d;
		} else
                    /* l holds the MAX value */
		    b[i1] = l;
	    } else if (u > l)
                 /* u holds the MAX value */
		b[i1] = u;
	    else
                /* l holds the MAX value */
		b[i1] = l;

            /* additional zero bounding for Smith Waterman. */
	    if (b[i1] < 0)
		b[i1] = 0;

            /*
             * Keep an idea of what the maximum score found in the SW matrix
             * is. Also keep its location.
             */
	    if (b[i1] > max_score) { 
		max_score = b[i1];
		c1 = i1;
		c2 = i2;
	    }
	}
    }

    free(lower);
    free(A);
    free(B);

    /* Coordinate in sequence 1 where the maximum score is found. */
    *end_coord1 = c1;
    /* Coordinate in sequence 2 where the maximum score is found. */
    *end_coord2 = c2;

    return max_score;
}

/**
 * Returns full smith-waterman alignment of the two sequences. 
 *
 * @param s1 is the null terminated string representing the first sequence
 *     to be aligned. <b>NB.</b> this is the sequence that forms the horizontal
 *     axis of the alignment matrix.
 * @param s2 is the null terminated string representing the second sequence
 *     to be aligned. <b>NB.</b> this is the sequence that forms the vertical
 *     axis of the alignment matrix.
 * @param len1 is the length if the sequence supplied in the parameter s1.
 * @param len2 is the length if the sequence supplied in the parameter s2.
 * @param match the the score that should be attributed to two exactly
 *     matching symbols drawn from s1 and s2.
 * @param mismatch is the mismatch penalty that should be associated with
 *     two mismatched symbols  from s1 and s2.
 * @param gap is the gap penalty associated with introducing a gap into
 *     either of the sequences.
 * @param neutral is the score attributed to a pair of matching 'n' 
 *     symbols in the two sequences.
 * @param offset is the offset into either s1 or s2 (if the offset is
 *     negative) that the banded alignment should be centered on.
 * @param window is the window size for banded alignment. The actual width
 *     of the region of the SW matrix explored by this routine is 2*window+1.
 *     A negative value for the window size signals that full alignment
 *     (without banding) should be carried out.
 *
 * @see smith_waterman_score2, print_alignment
 *
 * @returns an alignStruct which describes the alignment of the supplied
 *     sequences. This structure contains a set of segments that describe
 *     the (un-gapped) regions of the the two sequences that are conserved.
 *     It is possible to reconstruct the alignment by filling in the gaps
 *     between segments.
 */
alignStruct *
smith_waterman_alignment2(char *s1, char *s2, int len1, int len2,
                          int match, int mismatch, int gap, int neutral,
                          int offset, int window)
{
    int len;
    int **a;
    char **p;
    int *ibuf;
    char *cbuf;
    int i1, i2;
    int max_score = 0;
    int c1 = 0, c2 = 0;
    int m, n;
    int d, u, l;
    int start1, stop1;
    int start2, stop2;
    int x2;
    alignStruct *as = NULL;
    int *AA;
    char *PP;
    Seg *seg;
    int gapping;
    int width, bufsize;
    char *lower;
    char pad_char = '*';

    /* Single shot initalisation of the substitution matrix */
    if (!mat_init_ok)
	mat_init(match, mismatch, gap, neutral, pad_char);

    /* Take max of user supplied length and actual length */
    if (len1 > (m = strlen(s1)))
	len1 = m;

    /* Take max of user supplied length and actual length */
    if (len2 > (m = strlen(s2)))
	len2 = m;

    /*
     * Banded alignment. Flag for this happening is a value for window >= 0.
     *
     * Default is to cover the full length of sequence 2.
     *
     * start2/stop2 are the row index bounds for exploring the SW matrix.
     *     The value taken on by each of these is bounded by zero and the
     *     length of the second sequence.
     *
     * NB. For +ve offset this describes a band which starts at a point 
     * that is offset into sequence1. A -ve offset applies the offset to
     * sequence2.
     */
    if (window < 0) { /* Unbanded */
	start2 = 0;
	stop2 = len2 - 1;
    } else {
	start2 = -offset - window;
	if (start2 < 0)
	    start2 = 0;
	if (start2 >= len2)
	    start2 = len2 - 1;

	stop2 = len1 - offset + window;
	if (stop2 < 0)
	    stop2 = 0;
	if (stop2 >= len2)
	    stop2 = len2 - 1;
    }

    if(window < 0) { /* Unbanded */
        width = len1 + 1;
    } else {
        width = 2 * window + 5; /* Not clear why this is 5 instead of 1 */
    }

    /*
     * Adding a row of zeros to the matrix together with base index of zero
     * accounts for the +2 term in the following statement.
     */
    bufsize = width * (stop2 - start2 + 2);

    ibuf = (int *) calloc(bufsize, sizeof(int));
    cbuf = (char *) calloc(bufsize, sizeof(char));

    lower = (char *) downcase(strdup(s1));

    i2 = start2 - 1;
    if (window < 0) {
	start1 = 0;
	stop1 = len1 - 1;
    } else {
	start1 = i2 + offset - window;
	if (start1 < 0)
	    start1 = 0;
	if (start1 >= len1)
	    start1 = len1 - 1;
	stop1 = i2 + offset + window;
	if (stop1 < 0)
	    stop1 = 0;
	if (stop1 >= len1)
	    stop1 = len1 - 1;
    }

    /*
     * a - is a 2D array the SW matrix (or band of it). NB. The first index
     *     of the constructed 2D array is the row of the matrix and the second
     *     is the column. The matrix is made one row larger than neccessary
     *     to hold a row of zeros. (consequence of calloc allocating the
     *     store and row [start2 -1] never being modified.)
     * p - is a 2D matrix that shadows 'a' and is used to hold the transition
     *     information for subsequent playback of the alignment.
     */
    a = (int **) calloc(stop2 - start2 + 3, sizeof(int *)) - start2 + 1;
    a[start2 - 1] = &ibuf[0] - start1 + 2;

    p = (char **) calloc(stop2 - start2 + 3, sizeof(char *)) - start2 + 1;
    p[start2 - 1] = &cbuf[0] - start1 + 2;

    /*
     * Perform the walk down each of the rows of the SW matrix.
     */
    for (m = 1, i2 = start2; i2 <= stop2; m++, i2++) {
	if (window < 0) {
	    start1 = 0;
	    stop1 = len1 - 1;
	} else {
	    start1 = i2 + offset - window;
	    if (start1 < 0)
		start1 = 0;
	    if (start1 >= len1)
		start1 = len1 - 1;
	    stop1 = i2 + offset + window;
	    if (stop1 < 0)
		stop1 = 0;
	    if (stop1 >= len1)
		stop1 = len1 - 1;
	}
	a[i2] = &ibuf[m * width] - start1 + 2;
	p[i2] = &cbuf[m * width] - start1 + 2;

	x2 = tolower((int) s2[i2]);

        /*
         * Itterate over elements in sequence s1 in the range start1 to
         * stop1. This corresponds to traversing a row in the SW matrix.
         */
	for (i1 = start1; i1 <= stop1; i1++) {

           /*
            * Compute the components of the max() term of the SW recurrence.
            *     d - is the diagonal (consume 1 symbol of each sequence).
            *     u - up
            *     l - left
            *
            *     AA is used as a surrogate for the cell of the SW matrix
            *        that is currently being computed.
            *     PP is used as a surrogate for the parallel shadow matrix
            *        that contains the trace-back (i.e. the actual route
            *        that was taken in the MAX() evaluation) for the current
            *        cell in the SW matrix.
            */

	    d = a[i2 - 1][i1 - 1] + sim_mat[(int) lower[i1]][x2];
	    u = a[i2 - 1][i1] + gap;
	    l = a[i2][i1 - 1] + gap;

	    AA = &a[i2][i1]; /* Alias for the cell in the SW matrix that we
                                are currently computing. */
	    PP = &p[i2][i1]; /* Same as above but for transition info. */

	    if (d >= u) {
		if (d >= l) {
		    *AA = d;
		    *PP = PATH_DIAGONAL;
		} else {
		    *AA = l;
		    *PP = PATH_LEFT;
		}
	    } else if (u > l) {
		*AA = u;
		*PP = PATH_UP;
	    } else {
		*AA = l;
		*PP = PATH_LEFT;
	    }
	    if (*AA <= 0) {
		*AA = 0;
		*PP = PATH_END;
	    } else if (*AA > max_score) {
		max_score = *AA;
		c1 = i1;
		c2 = i2;
	    }
	}
    }

    /*
     * Decant the data computed above into an alignment structure for return
     * to the user of this routine.
     *
     * Walks back from the maximum score it found on the matrix fill stage
     * above. (passed in c1, c2)
     * Depending upon what it finds in the saved state table 'p' it
     * looks at the next location in this table. It terminates when it finds
     * the PATH_END instruction (as inserted by the matrix fill stage).
     */
    if (max_score > 0) { /* i.e. found an alignment */

        /* Dyn-allocate the return structure */
	as = (alignStruct *) calloc(1, sizeof(alignStruct));
	as->score = max_score;

	as->end1 = c1; /* end location in sequence 1 */
	as->end2 = c2; /* end location in sequence 2 */

	len = 0; /* Length of alignment */
	gapping = 0; /* Flag used to allow accurate counting of gapped
                        segments in the alignment. */

	while ((m = p[c2][c1]) != PATH_END) {
	    if (m == PATH_DIAGONAL) {
		c2--;
		c1--;
		len++;
		gapping = 0; /* Count any future gap insertions as the start
                                of a new segment */
	    } else if (m == PATH_LEFT) {
		c1--;

		if (!gapping)
		    as->segs++;
		gapping = 1; /* Don't count as a new segment until we make the
                                next diagonal transition. */

	    } else if (m == PATH_UP) {
		c2--;

		if (!gapping)
		    as->segs++;
		gapping = 1;  /* Don't count as a new segment until we make the
                                 next diagonal transition. */
	    }
	}
	as->start1 = c1 + 1; /* Where the alignment starts in sequence 1 */
	as->start2 = c2 + 1; /* Where the alignment starts in sequence 2 */

	if (len)
	    as->segs++;

	as->len = len;

	/*
         * NB. must be callocd. Following code assumes that elements of
         * structures are initalised to zero.
         */
	as->seg = (Seg *) calloc(as->segs, sizeof(Seg));

	/*
         * Works back from the last segment. NB. This actually points beyond
         * the end of the allocated space. However the code that follows makes
         * a decrement before attempting to dereference 'seg'. If any changes
         * are made to the code that follows this must be maintained.
         */
	seg = &as->seg[as->segs];

	gapping = 1;

        /*
         * Start a second walkback through the transtitions matrix.
         * Starts at the location of the global maximum in the SW matrix.
         */
	c1 = as->end1;
	c2 = as->end2;

	while ((m = p[c2][c1]) != PATH_END) {
	    if (m == PATH_DIAGONAL) {
		if (gapping) {
                    /*
                     * Called on first element and subsequent points
                     * when a gap has been inserted in either sequence.
                     */
		    gapping = 0;
		    seg--; /* See comment above with regard to ptr validity */
		    seg->end2 = c2;
		    seg->end1 = c1;
		}
		seg->start2 = c2;
		seg->start1 = c1;
		n = sim_mat[tolower((int) s1[c1])][tolower((int) s2[c2])];
		seg->score += n;
		seg->matches += (n > 0);
		c2--;
		c1--;
	    } else if (m == PATH_LEFT) {
		c1--;
		gapping = 1;
	    } else if (m == PATH_UP) {
		c2--;
		gapping = 1;
	    }
	}
    }
    free(a + start2 - 1);
    free(p + start2 - 1);
    free(ibuf);
    free(cbuf);
    free(lower);

    return as;
}

/**
 * Returns a value depending upon whether the two supplied base codes are
 * to be considered identical. This routine wraps the logic for dealing 
 * with pads, gaps and n's.
 *
 * @param c1 is a character representing a base code
 * @param c2 is a character representing a base code
 * @param match is the value that should be returned is c1 is identical to
 *        c2 and they are not in the set {'*', 'n', '-'}
 * @param mismatch is the value that should be returned if the two base codes
 *        are not identical and that neither supplied code is a '*', 'n'
 *        or '-'
 * @param gap is the value that should be returned if either (but not both)
 *        of the base codes are equal to '*'
 *
 * @returns
 */
int
equivalent_bases(int c1, int c2, int match, int mismatch, int gap)
{
    if (c1 == c2) {
	if (c1 != '*' && c1 != 'n' && c1 != '-')
	    return match;
	else
	    return 0;
    } else {
	if (c1 == '*' || c2 == '*')
	    return gap;
	else if (c1 == 'n' || c1 == '-' || c2 == 'n' || c2 == '-')
	    return 0;
	else
	    return mismatch;
    }
}

/**
 * This routine initalises a substitution matrix for use in Smith Waterman
 * and Needleman alignment algorithims. It depends upon a global 256 x 256
 * array of chars existing (sim_mat). The routine is some what wastefull
 * since only a small proprtion of the array will ever be accessed.
 * 
 * @param match is the score that should be attributed to a pair of matching
 *     bases or residue codes.
 * @param mismatch is the score that should be attributed to a pair of
 *     non-matching bases or residue codes.
 * @param gap is the penalty value for introducing a gap into either 
 *     sequence.
 * @param neutral is the value associated with a pair of 'n' values being
 *     matched.
 * @param pad_char is the character representing a pad in either of the
 *     sequences. The value that is associated with this is supplied
 *     by the parameter gap.
 */
void
mat_init(int match, int mismatch, int gap, int neutral, char pad_char)
{
    int c1, c2;

    for (c1 = 0; c1 < 256; c1++)
	for (c2 = 0; c2 < 256; c2++) {
	    if (c1 == c2) {
		if (c1 != '*' && c1 != 'n' && c1 != '-')
		    sim_mat[c1][c2] = match;
		else
		    sim_mat[c1][c2] = 0;
	    } else {
		if (c1 == pad_char || c2 == pad_char)
		    sim_mat[c1][c2] = sim_mat[c2][c1] = gap;
/*          else if( c1 == 'n' || c1 == '-' || c2 == 'n' || c2 == '-' )  */
		else if (c1 == 'n' || c2 == 'n')
		    sim_mat[c1][c2] = sim_mat[c2][c1] = neutral;
		else
		    sim_mat[c1][c2] = sim_mat[c2][c1] = mismatch;
	    }
	}
    for (c1 = 0; c1 < 256; c1++) {
	c2 = tolower(c1);
	sim_mat[c1][c2] = sim_mat[c1][c1];
	sim_mat[c2][c1] = sim_mat[c1][c1];
    }

    mat_init_ok = 1;
}

/**
 * Release storage associated with an alignment structure. This also frees
 * the internal data for the structure.
 *
 * @param as is the pointer to the alignment structure that was previously
 *    allocated on the heap.
 */
void
free_alignStruct(alignStruct *as)
{
    if (as) {
	if (as->seg)
	    free(as->seg);
	free(as);
    }
}

/**
 * Renders the supplied sequence alignment structure into a human readable
 * form. The output is as per the following example:
 *
 * e.g. seq1="AAAACTGAGGGATTCAAAA" and seq2="TTAACTGAGGATTCATAA" and the
 * parameters width=8, label1="seq1", label2="seq2" the following output
 * is generated:
 *
 *    2 AACTGAGG      9 seq1
 *    2 AACTGA-G      8 seq2
 *
 *  10 GATTCA     15 seq1
 *   9 GATTCA     14 seq2
 *
 * Here the numbers indicate the positions (starting offset=0) within
 * the original sequences that the symbols lie. So the "AA" patter lies
 * at position 3 of both sequences. Note that the width limits the
 * number of symbols that will be printed on a line before starting a
 * new block.
 *
 * @see smith_waterman_alignment2
 *
 * @param fp is the FILE stream to write the alignment rendition into.
 * @param s1 is the string representing the first sequence being aligned.
 *        This must <b>exactly</b> match the sequence provided to the
 *        alignment routine.
 * @param s2  is the string representing the second sequence being aligned.
 *        This must <b>exactly</b> match the sequence provided to the
 *        alignment routine.
 * @param as is the alignment data as returned by smith_waterman_alignment2().
 * @param width is the 'page width' that this routine should use to decide
 *        upon where to insert line breaks (on long sequences).
 * @param label1 is the string label that should be attached to sequence s1.
 *        This parameter may be null, in which case no label is attached to
 *        the sequence data.
 * @param label2 is the string label that should be attached to sequence s2.
 *        This parameter may be null, in which case no label is attached to
 *        the sequence data.
 */
void
print_alignment(FILE *fp, char *s1, char *s2, alignStruct *as,
                int width, char *label1, char *label2)
{
    int i, j1, j2, len = 0;
    char *buf1, *b1;
    char *buf2, *b2;
    Seg *seg, *pseg;
    int start1, end1, start2, end2;
    char padding = '-';

    if (fp && as && as->segs) { /* Sanity check on supplied parameters */

	for (i = 0; i < as->segs; i++) {
	    /*
             * Sanity check that all segments provided in the alignment
             * structure have matching lengths (bit of seq1 is the same
             * length as the corresponding bit of seq2).
             */
	    seg = &as->seg[i];
	    if (seg->end2 - seg->start2 != seg->end1 - seg->start1) {
		fprintf(stderr, "ERROR segment mismatch %d %d   %d %d\n",
                            seg->start1, seg->end1, seg->start2, seg->end2);
		exit(1);
	    }
	}

	/*
         * Compute the length of the string required to hold the aligned
         * version of the string. Note this also reserves space for the
         * padding characters.
         */
	seg = &as->seg[0];
	len = seg->end1 - seg->start1 + 1;

	/* Add the length of each subsequent segment together with space
           for pad chars. */
	for (i = 1; i < as->segs; i++) {
	    pseg = seg;
	    seg = &as->seg[i];
	    len += seg->end1 - seg->start1 + 1; /* segment length */
	    len += seg->start1 - pseg->end1 - 1 +
                       seg->start2 - pseg->end2 - 1; /* pad chars */
	}

	len++; /* space for null terminator */
	b1 = buf1 = (char *) calloc(len, sizeof(char));
	b2 = buf2 = (char *) calloc(len, sizeof(char));

	/*
         * Copy the sequence data into the new strings. This only does the
         * first segment.
         */
	seg = &as->seg[0];
	len = seg->start2 - seg->start1 + 1;
	for (j1 = seg->start1, j2 = seg->start2; j1 <= seg->end1; j1++, j2++, b1++, b2++) {
	    *b1 = s1[j1];
	    *b2 = s2[j2];
	}

	/*
         * Subsequent segments and introduce padding characters to fill
         * the gaps between the segments.
         */
	for (i = 1; i < as->segs; i++) {
	    pseg = seg;
	    seg = &as->seg[i];
	    if (pseg->end1 + 1 < seg->start1)
		for (j1 = pseg->end1 + 1, j2 = pseg->end2 + 1; j1 < seg->start1; j1++, j2++, b1++, b2++) {
		    *b2 = padding;
		    *b1 = s1[j1];
	    } else
		for (j2 = pseg->end2 + 1, j1 = pseg->end1 + 1; j2 < seg->start2; j1++, j2++, b1++, b2++) {
		    *b1 = padding;
		    *b2 = s2[j2];
		}

	    for (j1 = seg->start1, j2 = seg->start2; j1 <= seg->end1; j1++, j2++, b1++, b2++) {
		*b1 = s1[j1];
		*b2 = s2[j2];
	    }
	}

	/*
         * Do the actual output.
         */
	b1 = buf1;
	b2 = buf2;
	seg = &as->seg[0];
	start1 = end1 = seg->start1;
	start2 = end2 = seg->start2;
	while (*b1 && *b2) {
	    fprintf(fp, "%6d ", start1);
	    for (j1 = 0; j1 < width && *b1; j1++, b1++) {
		putc(*b1, fp);
		if (*b1 != padding)
		    end1++;
	    }
	    fprintf(fp, " %6d %s\n", end1 - 1, (label1 ? label1 : ""));

	    fprintf(fp, "%6d ", start2);
	    for (j2 = 0; j2 < width && *b2; j2++, b2++) {
		putc(*b2, fp);
		if (*b2 != padding)
		    end2++;
	    }
	    fprintf(fp, " %6d %s\n\n", end2 - 1, (label2 ? label2 : ""));
	    start1 = end1;
	    start2 = end2;

	}

	free(buf1);
	free(buf2);
    }
}
