/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

#include <stdlib.h>

#include "trace_align.h"

static const char *cvsid = "$Id: eval_local_minima.c 10638 2003-04-15 09:15:12Z dgm $";

/**
 * Determine the location of local "minima". Also computes a score based
 * on the change of gradient in the local region of the turning point. 
 *
 * Criterion for a local minimum is:
 *
 * (a) true local minimum:   f'[c-1] < 0 && f'[c+1] > 0
 *
 * (b) left shoulder:        f'[c] >= 0 && f'[c-1] > f'[c] && f'[c+1] > f'[c]
 *
 * (c) right shoulder        f'[c] <= 0 && f'[c-1] < f'[c] && f'[c+1] < f'[c]
 *
 * score at coord c is 
 *
 * max { |f'(x) - f'(y)| }
 * x <= c
 * y >= c
 * c-x < window
 * y-c < window
 *
 * where window is 5 samples
 *
 * @param samples is the count of sample points in the data parameter.
 * @param data is the collection of trace samples. First dimension is the
 *        channel, the second is the sequential trace index.
 * @param info is of the same dimension as data and contains the scoring
 *        for each of the transtitions in the in the original trace data.
 *        This is 'thinned' to have a value which marks the first occurence
 *        of the largest 'transition'. This corresponds to the W(X,X,t) on
 *        page 93 of the supporing paper.
 * @param left_side is of the same dimension as data and contains the
 *        'largest' change in the derivative of the corresponding 'channel'
 *        in the region to the left of any point in the original data. This
 *        corresponds to the 'L' term in the supporting paper.
 * @param right_side is of the same dimension as data and contains the
 *        'largest' change in the derivative of the corresponding 'channel'
 *        in the region to the right of any point in the original data. This
 *        corresponds to the 'R' term in the supporting paper.
 *
 * @see "Trace alignment and some of its applications", Bioinformatics,
 *      Vol. 14, no. 1, 1998, pp 92-97.
 */
void
eval_local_minima(int samples, int **data, int **info, int **left_side,
		  int **right_side)
{

    int i, c, n, x, dc, dleft, dright, max;
    int *d = (int *) calloc(samples, sizeof(int));
    int window = 5;

    for (i = base_a; i <= base_n; i++)
	for (c = 0; c < samples; c++)
	    info[i][c] = left_side[i][c] = right_side[i][c] = 0;

    for (i = base_a; i <= base_n; i++) {
	/* calculate the derivative on a channel by channel basis */

	for (c = 1; c < samples; c++)
	    d[c] = data[i][c] - data[i][c - 1];		/* simple gradient */

	for (c = window; c < samples - window; c++) {
	    dc = dleft = dright = 0;
	    if ((d[c - 1] < 0 && d[c + 1] > 0) ||	/* local minimum */
		((d[c] >= 0) && (d[c - 1] >= d[c]) && (d[c + 1] >= d[c])) ||	/* left shoulder */
		((d[c] <= 0) && (d[c - 1] <= d[c]) && (d[c + 1] <= d[c]))	/* right shoulder */
		) {

		/* estimate the derivative at c. 
		   This can be d[c] or 0 if at a local minimum */

		if (d[c - 1] < 0 && d[c + 1] > 0)
		    dc = 0;
		else
		    dc = d[c];


		/* left side */
		/* move through zero derivative region flanking c, if it exists */

		x = c - 1;
		while ((d[x] == 0) && (x > c - window))
		    x--;
		dleft = d[x];

		/*
		 *  move left along trace so long as:
		 *
		 * (a) sign of derivative does not change
		 * (b) absolute value of derivative increases
		 * (c) inside window 
		 *
		 */

		while ((x > c - window) && (dleft * d[x] > 0)
		       && (abs(d[x]) >= abs(d[x + 1])))
		    x--;

		left_side[i][c] = abs(d[x + 1] - dc);

		/* ditto for right side */
		/* move through zero derivative region flanking c, if it exists */

		x = c + 1;
		while ((d[x] == 0) && (x < c + window))
		    x++;
		dright = d[x];

		while ((x < c + window) && (dright * d[x] > 0)
		       && (abs(d[x]) >= abs(d[x - 1])))
		    x++;

		right_side[i][c] = abs(d[x - 1] - dc);

		if (dleft > 0 && dright < 0)	/* error ! we have found a flat-topped local max, so set to 0 */
		    left_side[i][c] = right_side[i][c] = 0;

		if (left_side[i][c] > 0 && right_side[i][c] > 0)
		    info[i][c] = left_side[i][c] + right_side[i][c];
	    }
	}
	/* reduce  runs of non-zero values to their max */

	for (c = window; c < samples - window; c++) {
	    if (info[i][c] > 0) {
		x = c;
		max = 0;
		while (x < samples && info[i][x] > 0) {
		    if (info[i][x] > max)
			max = info[i][x];
		    x++;
		}
		x = c;
		n = 0;
		while (x < samples && info[i][x] > 0) {
		    if (info[i][x] == max) {
			if (n > 0)
			    info[i][x] = left_side[i][x] = right_side[i][x] = 0;
			n++;
		    } else if (info[i][x] < max)
			info[i][x] = left_side[i][x] = right_side[i][x] = 0;
		    x++;
		}
	    }
	}
    }

    free(d);
}
