/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

#include <stdio.h>
#include <stdlib.h>

#include "trace_align.h"
#include <pair_memory.h>

static const char *cvsid = "$Id: gapped_trace_align.c 18816 2006-08-03 16:39:42Z rmd $";

/**
 * Perform a gapped trace alignment.
 *
 * @param samples is the size of trace_X arrays
 * @param trace_weight[A,C,G,T][i] is the array of weights for each base
 *        at trace coordinate i;
 * @param transition[i] is the weight for making a transition
 *        at trace coordinate i;
 * @param bases is the size of base array (# of bases)
 * @param base is the array of bases called [ indexed as 0,1,2,3,(4=N) ]
 * @param minimum is the bound on allowed length of time before changing state.
 * @param search_mode is the SMITH_WATERMAN best local similarity
 * @param trace_coord is the array giving coords for transitions
 *        (pre-allocated, of size bases+1) 
 * @param align_score is the array giving score for each base
 *        (pre-allocated) backtracking is performed via a path[b][c]
 *        matrix which is set_bit-packed (2 bits per cell) this reduces
 *        memory by a factor of 4 note paths can go left or diagonal or up.
 */
void
gapped_trace_align(int samples,
		   int **trace_weight,
		   int **local_min,
		   int bases, char *base,
		   int minimum,
		   int window_band,
		   int gap_penalty,
		   int *trace_coord,
		   int *align_score,
		   int *total_score,
		   int *base_start,
		   int *base_end, int *trace_start, int *trace_end)
{

    unsigned char **path, *P;
    int b, c, m = 0, i;
    int *score1, *score2, *s1, *s2, *s3;
    int change_state, no_change, max_c = 0, max_b = 0, max_score = INT_MIN;
    int score, last, lastc, in_gap;
    int *w;
    int *wc, **cum = NULL;
    int samples8;
    int dbit[4] =
    {1, 4, 16, 64};		/* for bit packing */
    int vbit[4] =
    {2, 8, 32, 128};
    int hbit[4] =
    {3, 12, 48, 192};
    int previous_state, current_state, vertical_gap;
    int *loc_min;
    int *best_col, *best_base;	/* for vertical gaps */

    path = (unsigned char **) calloc(bases + 1, sizeof(unsigned char *));
    path++;

    samples8 = 2 + samples / 4;	/* for bit packing */

    for (b = -1; b < bases; b++) {
	path[b] = (unsigned char *) calloc(samples8, sizeof(unsigned char));
	path[b]++;
    }


    score1 = (int *) calloc(samples + 1, sizeof(int));
    score2 = (int *) calloc(samples + 1, sizeof(int));

    s1 = score1 + 1;
    s2 = score2 + 1;

    best_col = (int *) calloc(samples + 1, sizeof(int));
    best_base = (int *) calloc(samples + 1, sizeof(int));

    cum = (int **) calloc(5, sizeof(int *));
    for (i = 0; i < 5; i++) {
	cum[i] = (int *) calloc(samples + 1, sizeof(int)) + 1;
	cum[i][0] = trace_weight[i][0];
    }

    for (c = 1; c < samples; c++) {
	for (i = 0; i < 5; i++)
	    cum[i][c] = cum[i][c - 1] + trace_weight[i][c];
    }

    previous_state = base_n;
    for (b = 0; b < bases; b++) {
	s3 = s1;
	s1 = s2;
	s2 = s3;
	P = path[b];

	current_state = base[b];
	w = trace_weight[(int) base[b]];
	wc = cum[previous_state];


	if (previous_state == current_state)
	    loc_min = local_min[(int) base[b]];
	else
	    loc_min = local_min[(int) base_n];

	for (c = 0; c < minimum; c++) {
	    no_change = s1[c - 1] + w[c] - loc_min[c];
	    change_state = s2[0] + w[c] + wc[c - 1] + loc_min[c];
	    vertical_gap = best_col[c] - gap_penalty + loc_min[c];
	    if (vertical_gap > no_change && vertical_gap > change_state) {
		s1[c] = vertical_gap;
		P[c / 4] |= vbit[c % 4];
		if (s1[c] > max_score) {
		    max_score = s1[c];
		    max_c = c;
		    max_b = b;
		}
	    } else if ((no_change > change_state) && (no_change > 0)) {		/* left */
		s1[c] = no_change;
		P[c / 4] |= hbit[c % 4];
		if (s1[c] > max_score) {
		    max_score = s1[c];
		    max_c = c;
		    max_b = b;
		}
	    } else if (change_state < 0) {	/* terminate */
		s1[c] = 0;
	    } else {		/* diagonal */
		P[c / 4] |= dbit[c % 4];	/* 2-bit packing */
		s1[c] = change_state;
		if (s1[c] > max_score) {
		    max_score = s1[c];
		    max_c = c;
		    max_b = b;
		}
	    }

	    if (best_col[c] < s1[c]) {
		do_not_forget(c, b);
		best_col[c] = s1[c];
		best_base[c] = b;
	    }
	}

	for (c = minimum; c < samples; c++) {
	    no_change = s1[c - 1] + w[c] - loc_min[c];
	    if (no_change < 0)
		no_change = 0;
	    change_state =
		s2[c - minimum] + w[c] + wc[c - 1] - wc[c - minimum] +
		loc_min[c];
	    if (change_state < 0)
		change_state = 0;

	    vertical_gap =
		best_col[c - minimum] - gap_penalty + w[c] + wc[c - 1] - wc[c -
								      minimum]
		+ loc_min[c];

	    if (vertical_gap > no_change && vertical_gap > change_state) {
		s1[c] = vertical_gap;
		P[c / 4] |= vbit[c % 4];
		if (s1[c] > max_score) {
		    max_score = s1[c];
		    max_c = c;
		    max_b = b;
		}
	    } else if ((no_change > change_state)) {	/* left */
		s1[c] = no_change;
		P[c / 4] |= hbit[c % 4];
		if (s1[c] > max_score) {
		    max_score = s1[c];
		    max_c = c;
		    max_b = b;
		}
	    } else if (change_state == 0) {
		s1[c] = 0;
	    } else {		/* diagonal */
		P[c / 4] |= dbit[c % 4];	/* bit packing */
		s1[c] = change_state;
		if (s1[c] > max_score) {
		    max_score = s1[c];
		    max_c = c;
		    max_b = b;
		}
	    }

	    if (best_col[c] < s1[c]) {
		do_not_forget(c, b);
		best_col[c] = s1[c];
		best_base[c] = b;
	    }
	}
	previous_state = current_state;
    }


    printf("!max score: %d minimum %d max_c: %d max_b: %d bases: %d\n",
	   max_score, minimum, max_c, max_b, bases);

    *total_score = max_score;
    *base_end = max_b;
    *trace_end = max_c;

    for (b = 0; b < bases; b++)
	trace_coord[b] = -1;

    trace_coord[max_b] = max_c;

    b = max_b;
    c = max_c;
    score = 0;
    last = 0;
    lastc = c;
    vertical_gap = 0;
    in_gap = 0;

    while (b >= 0 && c >= 0) {
	if ((path[b][c / 4] & hbit[c % 4]) == vbit[c % 4]) {	/* vertical */
	    int new_b;

	    align_score[b] = 0;
	    trace_coord[b] = c;
	    last = score;
	    lastc = c;
	    new_b = remember(c, b);
	    while (b >= new_b)
		trace_coord[b--] = c;

	    b = new_b;
	    if (!in_gap)
		vertical_gap =
		    -gap_penalty + cum[(int) base[b]][c - 1] -
		    cum[(int) base[b]][c - m];
	    in_gap = 1;
	} else if ((path[b][c / 4] & hbit[c % 4]) == dbit[c % 4]) {	/* diagonal */
	    score += trace_weight[(int) base[b]][c] + vertical_gap;
	    vertical_gap = 0;
	    in_gap = 0;
	    align_score[b] = score - last;
	    last = score;
	    lastc = c;

	    if (base[b] == base[b - 1])
		score += local_min[base_n][c];
	    else
		score += local_min[(int) base[b]][c];

	    vertical_gap = 0;
	    trace_coord[b] = c;
	    if (!minimum) {
		b--;
		c--;
	    } else {
		b--;
		if (minimum <= c)
		    m = minimum;
		else
		    m = c;
		if (b >= 0)
		    score +=
			cum[(int) base[b]][c - 1] - cum[(int) base[b]][c - m];

		c -= m;
	    }
	} else if ((path[b][c / 4] & hbit[c % 4]) == hbit[c % 4]) {	/* horizontal */
	    score += trace_weight[(int) base[b]][c] + vertical_gap;
	    if (base[b] == base[b - 1])
		score -= local_min[(int) base[b]][c];
	    else
		score -= local_min[base_n][c];

	    vertical_gap = 0;
	    in_gap = 0;
	    c--;
	} else {		/* stop */
	    break;
	}
    }

    printf("backtrack score %d\n", score);

    *base_start = b;
    *trace_start = c;

    free(score1);
    free(score2);

    for (b = -1; b < bases; b++) {
	free(path[b] - 1);
    }

    free(path - 1);

    for (i = 0; i < 5; i++)
	free(cum[i] - 1);
    free(cum);
}
