/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

#include <stdio.h>
#include <stdlib.h>

#include <io_lib/scf.h>
#include "trace_align.h"

static const char *cvsid = "$Id: trace_align.c 18816 2006-08-03 16:39:42Z rmd $";

/**
 * Perform trace alignment. This is a 2 stage process with this routine
 * performing pre-processing and then handing control to either
 * gapped_trace_align or trace_align2 depending upon the value of the
 * params->gap_penalty. A gap penalty > 0 indicates that a gapped alignment
 * should be carried out.
 *
 * @param scf is the trace data stored in a Scf structure
 * @param trace_start is the start coord in trace
 * @param trace_end is the end coord in trace
 * @param seq is the sequence to align
 * @param seq_start is the start coord in seq
 * @param seq_end is the end coord in seq
 * @param params is the set of alignment parameters
 * @param rescore is a flag indicating that this routine should rescore
 *        the alignment and return a pointer to a ReScore struct.
 * @param trace_coord is the pre-allocated array returning the trace
 *        coordinates at which state transitions occur
 * @param align_score is the pre-allocated array returning the score
 *        for each aligned base
 * @param total_score is the return parameter for the alignment score 
 * @param base_start is the return parameter for the start of the alignment
 *        in the seq
 * @param base_end is the return parameter for the end of the alignment
 *        in the seq
 * @param t_start is the return parameter for the start of the alignment
 *        in the trace
 * @param t_end is the return parameter for the end of the alignment in
 *        the trace
 *
 * @returns a ReScore structure indicating the alignment that has been
 *          achieved.
 *
 * @see gapped_trace_align, trace_align2
 */

ReScore *
trace_align(Scf * scf, int trace_start, int trace_end, char *seq, int seq_start,
	    int seq_end, t_params * params, int rescore, int *trace_coord,
	    int *align_score, int *total_score, int *base_start, int *base_end,
	    int *t_start, int *t_end)
{				/* the end of the alignment in the trace */
    /* NOTE ON COORDINATE SYSTEM:
       Bases are numbered from position 0. 

       The active region of the sequence is [seq_start,seq_end]

       The arrays seq, trace_coord, w_score MUST appear to be allocated from 0
       to at least seq_end, since the data are written back assuming this
       (you can fake it by allocating a block P of size seq_end-seq_start+1
       and then passing in a pointer to P-seq_start)

     */
    /* preprocess the data */

    int **trace_weight, **local_min, **left_min, **right_min;
    int **sample_data;
    int samples;
    int bases;
    char *base;
    int *max_trace, *max_trace2, floor_level;
    Header *h = &scf->header;
    int i, c, n, b;
    ReScore *re_score = NULL;

    int min = params->minstate;	/* min no of trace coords spent in each state */
    int search_mode = params->mode;	/* either SMITH_WATERMAN or NEEDLEMAN */
    int samples_per_base = params->samples_per_base;	/* window for banded alignement */
    float floor = params->floor;	/* number between 0 and 1 subtracted from traces */

    if (trace_start < 0 || trace_end < 0 || trace_end < trace_start
	|| trace_end >= h->samples) {
	trace_start = 0;
	trace_end = h->samples - 1;
    }
    samples = trace_end - trace_start + 1;

    if (seq_start < 0 || seq_end < 0 || seq_end < seq_start) {
	seq_start = 0;
	seq_end = h->bases - 1;
    }
    bases = seq_end - seq_start + 1;


    /*  printf ("rescore %d\n", rescore ); */

    /* arrays for local min info */

    trace_weight = alloc_sample_matrix(samples);
    local_min = alloc_sample_matrix(samples);
    left_min = alloc_sample_matrix(samples);
    right_min = alloc_sample_matrix(samples);

    /* copy the scf data into something easier to handle */

    sample_data = make_sample_data(scf, trace_start, samples);

    max_trace = sample_data[base_n];	/* note aliasing */
    max_trace2 = (int *) calloc(samples, sizeof(int));


    base = (char *) calloc(bases + 2, sizeof(char)) + 1;

    /* create numeric codes for bases. 
       NOTE anything not a,c,g,t is coded as n */

    base[-1] = base_n;
    for (i = 0; i < bases; i++) {
	base[i] = i_base(seq[seq_start + i]);
	if (base[i] > base_n)
	    base[i] = base_n;
    }

    /*
     * compute the envelope of traces. Note that
     * max_trace = sample_data[base_n]. This looks like it places the
     * envelope trace in the 'N' channel of the data.
     */

    trace_envelope(samples, sample_data, max_trace, max_trace2);

    /*
     * subtract 90% (or whatever floor is set to) of the max trace
     * from each trace to make the weights mostly negative, for
     * Smith-Waterman style alignment
     */

    for (c = 0; c < samples; c++) {

	floor_level = floor * max_trace[c];

	for (i = base_a; i <= base_t; i++)
	    trace_weight[i][c] = (int) sample_data[i][c] - floor_level;

	trace_weight[base_n][c] = 0;
    }

    /* create the data on local minima, shoulders etc */

    eval_local_minima(samples, sample_data, local_min, left_min, right_min);

    /* do the alignment */

    if (params->gap_penalty <= 0) {
	trace_align2(samples, trace_weight, local_min, bases, base, min,
		     search_mode, samples_per_base, &trace_coord[seq_start],
		   &align_score[seq_start], total_score, base_start, base_end,
		     t_start, t_end);
    } else {			/* gapped alignment */
	gapped_trace_align(samples, trace_weight, local_min, bases, base, min,
			   samples_per_base, params->gap_penalty,
			   &trace_coord[seq_start], &align_score[seq_start],
			   total_score, base_start, base_end, t_start, t_end);
    }

    /*  printf("! score: %d base_start %d base_end %d\n", *total_score, *base_start, *base_end ); */

    *base_start += seq_start;
    *base_end += seq_start;

    if (rescore) {		/* create and fill a ReScore structure describing the alignment in detail */
	int bb;

	/* rescore the alignment */
	re_score = allocReScore(*base_start, *base_end);

	/*      for(c=0;c<samples;c++)
	   printf("%5d %d\n", c, local_min[base_n][c] ); */

	/* compute the max-possible scorings */

	for (b = *base_start; b < *base_end; b++) {
	    /*      printf("%5d %c\n", b, seq[b]); */
	    bb = b - seq_start;
	    n = base[bb];
	    for (c = trace_coord[b] + 1; c <= trace_coord[b + 1]; c++) {
		re_score->max[b] += max_trace[c];	/* max possible score */
		re_score->score[b] += sample_data[n][c];	/* -max_trace2[c];  score for aligned base - height of second-max */
	    }
	    re_score->right[b] = left_min[base_n][trace_coord[b + 1]];	/* right-hand side of peak */
	    re_score->left[b] = right_min[base_n][trace_coord[b]];	/* left-hand side of peak */
	    re_score->width[b] = trace_coord[b + 1] - trace_coord[b];	/* width of aligned base */
	    for (c = trace_coord[b] + 2; c < trace_coord[b + 1] - 1; c++) {	/* note we ignore internal local minima if they are 2 pixels from the edge */
		re_score->internal[b] += local_min[base_n][c];	/* internal derivative info - the sum of all local min values inside for the maximum base - possibly CHANGE THIS!!!! */
	    }
	}
    }
    /* translate the trace coordinates back into the input space */

    for (i = *base_start; i <= *base_end; i++)
	trace_coord[i] += trace_start;

    *t_start += trace_start;
    *t_end += trace_start;

    free_sample_matrix(sample_data);
    free_sample_matrix(left_min);
    free_sample_matrix(right_min);
    free_sample_matrix(local_min);
    free_sample_matrix(trace_weight);

    free(base - 1);
    free(max_trace2);

    return re_score;
}
