/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

#include <stdio.h>
#include <stdlib.h>

#include "trace_align.h"

static const char *cvsid = "$Id: trace_align2.c 10638 2003-04-15 09:15:12Z dgm $";

/**
 * This routine performs the actual alignment of a search sequence and
 * a set of trace data. The algorithim used is described in the paper
 * listed below. Backtracking is performed via a path[b][c] matrix which is
 * set_bit-packed (1 bit per cell) this reduces memory by a factor of
 * 8 note paths can only go left or diagonal not up. The default mode of
 * operation (and hence the most tested flow control path) is NEEDLEMAN.
 *
 * The inputs to this routine are: samples, trace_weight, local_min, bases
 * base, minimum and search_mode. The output is trace_coord.
 *
 * @see "Trace alignment and some of its applications", Bioinformatics,
 *      Vol. 14, no. 1, 1998, pp 92-97.
 *
 * @param samples is the size of trace_X arrays
 * @param trace_weight is the 2D array([A,C,G,T][i]) of trace values that
 *        have typicaly been scaled to be 'mostly negative'
 * @param local_min is the array, corresponding to the 'W' on page 93 of
 *        supporting paper. It is of the same dimensionality as the trace
 *        data but has zeros everywhere except where a local minima occurs
 *        in the trace. At those locations there is a score which corresponds
 *        to the 'steepness' of the sides. A large score means a more pronounced
 *        feature in the data.
 * @param bases is the size of bases array (# of bases)
 * @param base is the array of bases [indexed as 0,1,2,3,(4=N)] representing
 *        the target pattern to be matched.
 * @param minimum is the bound on allowed length of time before changing state.
 * @param search_mode can be either SMITH_WATERMAN (best local similarity) or
 *        NEEDLEMAN (global similarity, in the sense that it stretches from the
 *        end of the sequence (not the trace). The default is NEEDLEMAN.
 * @param trace_coord is the array giving coords for transitions
 *        (pre-allocated, of size bases+1) w_score is the array giving
 *        score for each base (pre-allocated)
 * @param samples_per_base is the expected max no. of samples per base, in
 *        order to do a banded alignment set this to 0 for a full alignment.
 */
void
trace_align2(int samples,
	     int **trace_weight,
	     int **local_min,
	     int bases,
	     char *base,
	     int minimum,
	     int search_mode,
	     int samples_per_base,
	     int *trace_coord,
	     int *align_score,
	     int *total_score,
	     int *base_start,
	     int *base_end,
	     int *trace_start,
	     int *trace_end)
{

    unsigned char **path, *P;
    int b, c, m, i;
    int *score1, *score2, *s1, *s2, *s3;
    int change_state, no_change, max_c = 0, max_b = 0, max_score = INT_MIN;
    int score, last, lastc;
    int *w;
    int *wc, **cum = NULL;
    int samples8;
    int bit[8] =
    {1, 2, 4, 8, 16, 32, 64, 128};	/* for bit packing */
    int pat;
    int previous_state, current_state;
    int *loc_min;
    int win_start, win_end;
    float scale;
    int cells = 0;
    int window_band;

    /* the window band half-width */

    window_band = bases * samples_per_base;

#ifdef DEBUG
    /* store the scores to check backtracking */
    int **temp;
    int tdiff, t_last;
#endif

    path = (unsigned char **) calloc(bases + 1, sizeof(unsigned char *));
    path++;

    samples8 = 2 + samples / 8;	/* for bit packing */

    for (b = -1; b < bases; b++) {
	path[b] = (unsigned char *) calloc(samples8, sizeof(unsigned char));
	path[b]++;
    }


    score1 = (int *) calloc(samples + 1, sizeof(int));
    score2 = (int *) calloc(samples + 1, sizeof(int));

    s1 = score1 + 1;
    s2 = score2 + 1;

#ifdef DEBUG

    temp = (int **) calloc(bases + 2, sizeof(int *)) + 1;
    for (b = -1; b < bases; b++)
	temp[b] = (int *) calloc(samples + 2, sizeof(int)) + 1;
#endif

    cum = (int **) calloc(5, sizeof(int *));
    for (i = 0; i < 5; i++)
	cum[i] = (int *) calloc(samples + 1, sizeof(int)) + 1;

    cum[base_a][0] = trace_weight[base_a][0];
    cum[base_c][0] = trace_weight[base_c][0];
    cum[base_g][0] = trace_weight[base_g][0];
    cum[base_t][0] = trace_weight[base_t][0];
    cum[base_n][0] = trace_weight[base_n][0];

    for (c = 1; c < samples; c++) {
	cum[base_a][c] = cum[base_a][c - 1] + trace_weight[base_a][c];
	cum[base_c][c] = cum[base_c][c - 1] + trace_weight[base_c][c];
	cum[base_g][c] = cum[base_g][c - 1] + trace_weight[base_g][c];
	cum[base_t][c] = cum[base_t][c - 1] + trace_weight[base_t][c];
	cum[base_n][c] = cum[base_n][c - 1] + trace_weight[base_n][c];
    }

    scale = (samples / (bases + 1.0e-5));

    previous_state = base_n;
    for (b = 0; b < bases; b++) {
	s3 = s1;
	s1 = s2;
	s2 = s3;
	P = path[b];

	current_state = base[b];
	w = trace_weight[(int) base[b]];
	wc = cum[previous_state];


	if (previous_state == current_state)	/* the state transition penalty */
	    loc_min = local_min[(int) base[b]];
	else
	    loc_min = local_min[base_n];


	if (window_band > 0) {
	    win_start = b * scale - window_band;
	    if (win_start < 0)
		win_start = 0;
	    win_end = b * scale + window_band;
	    if (win_end > samples - 1)
		win_end = samples - 1;
	} else {
	    win_start = 0;
	    win_end = samples - 1;
	}

	if (search_mode == SMITH_WATERMAN) {

	    for (c = win_start; c < minimum; c++) {
		no_change = s1[c - 1] + w[c] - loc_min[c];
		change_state = s2[0] + w[c] + wc[c - 1] + loc_min[c];

		if ((no_change > change_state) && (no_change > 0)) {	/* left */
		    s1[c] = no_change;
		    if (s1[c] > max_score) {
			max_score = s1[c];
			max_c = c;
			max_b = b;
		    }
		} else if (change_state < 0) {
		    s1[c] = 0;
		} else {	/* diagonal */
		    P[c / 8] |= bit[c % 8];	/* bit packing */
		    s1[c] = change_state;
		    if (s1[c] > max_score) {
			max_score = s1[c];
			max_c = c;
			max_b = b;
		    }
		}

	    }

	    if (win_start < minimum)
		win_start = minimum;

	    for (c = win_start; c <= win_end; c++) {
		cells++;
		no_change = s1[c - 1] + w[c] - loc_min[c];
		if (no_change < 0)
		    no_change = 0;
		change_state =
		    s2[c - minimum] + w[c] + wc[c - 1] - wc[c - minimum] +
		    loc_min[c];
		if (change_state < 0)
		    change_state = 0;
		if ((no_change > change_state)) {	/* left */
		    s1[c] = no_change;
		    if (s1[c] > max_score) {
			max_score = s1[c];
			max_c = c;
			max_b = b;
		    }
		} else if (change_state == 0) {
		    s1[c] = 0;
		} else {	/* diagonal */
		    P[c / 8] |= bit[c % 8];	/* bit packing */
		    s1[c] = change_state;
		    if (s1[c] > max_score) {
			max_score = s1[c];
			max_c = c;
			max_b = b;
		    }
		}
	    }
	} else {
	    for (c = 0; c < minimum; c++) {
		no_change = s1[c - 1] - loc_min[c];
		change_state = s2[0] + wc[c - 1] + loc_min[c];

		if (change_state < no_change) {
		    s1[c] = no_change + w[c];
		} else {
		    P[c / 8] |= bit[c % 8];
		    s1[c] = change_state + w[c];
		}

	    }

	    for (c = minimum; c < samples; c++) {
		no_change = s1[c - 1] - loc_min[c];
		change_state =
		    s2[c - minimum] + wc[c - 1] - wc[c - minimum] + loc_min[c];

		if (change_state < no_change) {
		    s1[c] = no_change + w[c];
		} else {
		    P[c / 8] |= bit[c % 8];
		    s1[c] = change_state + w[c];
		}

	    }
	}
	previous_state = current_state;

#ifdef DEBUG
	for (c = -1; c < samples; c++)
	    temp[b][c] = s1[c];
#endif
    }

    if (search_mode != SMITH_WATERMAN) {
	max_score = s1[0];
	max_c = 0;
	max_b = bases - 1;
	for (c = 0; c < samples; c++) {
	    if (s1[c] > max_score) {
		max_score = s1[c];
		max_c = c;
	    }
	}
    }
    /*printf("\n!max score: %d minimum %d max_c: %d max_b: %d bases: %d mode:%d cells: %d window_band: %d\n",
	   max_score, minimum, max_c, max_b, bases, search_mode,
	   cells, window_band); */

    /* back-tracking phase */

    *total_score = max_score;
    *base_end = max_b;
    *trace_end = max_c;

    for (b = 0; b < bases; b++)
	trace_coord[b] = 0;

    trace_coord[max_b] = max_c;
    b = max_b;
    c = max_c;
    score = 0;
    last = 0;
    lastc = c;

    while (b >= 0 && c >= 0 && (search_mode == NEEDLEMAN || score <= max_score)) {
	pat = bit[c % 8] & path[b][c / 8];

	if (pat > 0) {


	    /*printf("%5d %5d %c %5d %10d %10d\n", b, c, c_base(base[b]), score-last,  local_min[base[b]][c], trace_weight[base[b]][c]);    */

	    align_score[b] = score - last;
	    last = score;
	    lastc = c;

	    if (base[b] == base[b - 1])
		score += local_min[(int) base[b]][c];
	    else
		score += local_min[base_n][c];

	    score += trace_weight[(int) base[b]][c];

	    trace_coord[b] = c;

	    if (!minimum) {
		b--;
		c--;
	    } else {
		b--;
		if (minimum <= c)
		    m = minimum;
		else
		    m = c;
		if (b >= 0 && base[b] <= base_n)
		    score +=
			cum[(int) base[b]][c - 1] - cum[(int) base[b]][c - m];
		else if (b >= 0)
		    score += cum[base_n][c - 1] - cum[base_n][c - m];
		c -= m;
	    }
	} else if (pat == 0) {
	    score += trace_weight[(int) base[b]][c];
	    c--;
	}
    } /* while */

    if (b < 0)
	b = 0;
    *base_start = b;
    if (c < 0)
	c = 0;
    *trace_start = c;

    free(score1);
    free(score2);

    for (b = -1; b < bases; b++) {
	free(path[b] - 1);
    }

    free(path - 1);

    for (i = 0; i < 5; i++)
	free(cum[i] - 1);
    free(cum);
}
