/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

/**
 * @file bases.c
 * @brief Two routines for mapping enum basecodes to chars and back.
 */

#include <ctype.h>

#include "bases.h"

static const char *cvsid = "$Id: bases.c 18818 2006-08-03 16:41:58Z rmd $";

/**
 * Convert a character code for a base into the enumeration value defined
 * in bases.h
 *
 * Performs a single-shot initalisation to build a lookup table which uses
 * the ordinal number of the base character into an enumeration value from
 * bases.h. Although not terrible this would be better implemented as a
 * map object (to reduce memory wastage slightly).
 *
 * @param base is the character representing the required base code.
 *
 * @returns one of base_a, base_c, base_g, base_t, base_n, base_i in normal
 * operation and -1 or base_o in error situations.
 */
int 
i_base(char base)
{

    static int ibase[BASES_TABLE_SIZE];
    static int init;

    if (!init) {
	int i;
	for (i = 0; i < BASES_TABLE_SIZE; i++)
	    ibase[i] = base_o;

	ibase['a'] = base_a;
	ibase['c'] = base_c;
	ibase['g'] = base_g;
	ibase['t'] = base_t;
	ibase['u'] = base_t;
	ibase['n'] = base_n;
	ibase['x'] = base_n;	/* x maps to n */

	ibase['*'] = base_n;	/* reverted */
	ibase['-'] = base_i;
	ibase['.'] = -1;

	init = 1;
    }
    return ibase[tolower((int) base)];
}

/**
 * Convert an enumeration value, representing a base, defined in bases.h
 * into a character representing the base. This represents (approximately)
 * the inverse of i_base. 
 *
 * Performs a single-shot initalisation to build a lookup table which uses
 * the enumeration values as the key and maps this to the associated character
 * representing the base. Although not terrible this would be better
 * implemented as a map object (to reduce memory wastage slightly).
 *
 * @param base is the enumeration value for a base (see bases.h)
 *
 * @returns one of 'a', 'c', 'g', 't', 'n', padding_char in normal
 * operation and '.' or ' ' in error situations.
 *
 * @see i_base
 */
char 
c_base(int base)
{

    static char cbase[BASES_TABLE_SIZE];
    static int init = 0;

    if (!init) {

	int i;
	for (i = 0; i < BASES_TABLE_SIZE; i++)
	    cbase[i] = '.';

	cbase[base_a] = 'a';
	cbase[base_c] = 'c';
	cbase[base_g] = 'g';
	cbase[base_t] = 't';
	cbase[base_n] = 'n';
	cbase[base_i] = PADDING_CHAR;

	init = 1;
    }
    if (base <= BASES_TABLE_SIZE && base >= 0)
	return cbase[base];
    else
	return ' ';
}
