/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

/**
 * @file pair_memory.c
 * Stuff for handling pairs of integers.
 * Needed to store pairs [row, col] so that we can retrieve the nearest
 * pair less than a given input pair [Row, Col] st Col == col and Row > row
 * and used for lin_align.c
 */

#include <stdio.h>
#include <stdlib.h>

#include "pair_memory.h"

static const char *cvsid = "$Id: pair_memory.c 10638 2003-04-15 09:15:12Z dgm $";


#define LIMIT_RPAIR_SIZE 10000

static RPAIR *rpair = NULL;
static int rpairs = 0;
static int rpair_size = 0;
static int rpairs_sorted = 0;
static int limit_rpair_size = LIMIT_RPAIR_SIZE;

/**
 * Compare the 'col' component of two RPAIR objects. If they are identical
 * then verify by checking that the 'row' components are also identical. 
 * Used as the sorting function for qsort. See function remember.
 *
 * @param a is the first RPAIR object.
 * @param b is the second RPAIR object.
 *
 * @returns 0 if the RPAIR's are equal in both their row and column values
 *   and non-zero if they differ.
 */
int
rpair_cmp(const void *a, const void *b)
{
    RPAIR *A = (RPAIR *) a;
    RPAIR *B = (RPAIR *) b;
    int n = A->col - B->col;

    if (n == 0)
	n = A->row - B->row;

    return n;
}

/**
 * Set the limit of the maximum number of pairs that can be stored. Since
 * this also frees the allocated storage it must be called early (and
 * not after any pairs have been stored).
 *
 * @param max_bytes is the maximum number of bytes that can be used by the
 * rpair's store. <b>NB</b>this is divided by the size of RPAIR.
 */
void
rpair_init(int max_bytes)
{
    limit_rpair_size = max_bytes / sizeof(RPAIR);
    free_rpairs();
}

/**
 * Free storage associated with the rpair's collection. This also initalises
 * internal variables to zero/NULL.
 */
void
free_rpairs()
{
    /*  fprintf(stderr,"FORGET: rpairs: %d\n", rpairs ); */
    if (rpair)
	free(rpair);
    rpair = NULL;
    rpair_size = 0;
    rpairs = 0;
    rpairs_sorted = 0;
}

/**
 * Store a pair object. This routine will allocate (and extend) its own
 * vector of RPAIR objects. If the user specified limit is exceeded then
 * this routine will return zero. The storage vector is enlarged by a
 * factor of two each time it is exhausted (upto the user defined limit).
 *
 * @param col is the column offset to store.
 * @param row is the row offset to store.
 *
 * @returns 1 on success and 0 otherwise.
 */
int
do_not_forget(int col, int row)
{

    if (rpairs > limit_rpair_size)
	return 0;		/* failure - ran out of memory */

    if (rpairs >= rpair_size) {
	rpair_size = (rpairs == 0 ? 10000 : 2 * rpairs);

	if (rpair_size > limit_rpair_size)	/* enforce the limit */
	    rpair_size = limit_rpair_size;

	rpair = (RPAIR *) realloc(rpair, sizeof(RPAIR) * rpair_size);
	if (rpair == NULL) {
	    fprintf(stderr, "ERROR: Memory allocation failure in do_not_forget(). rpairs: %d rpair_size: %d\n", rpairs, rpair_size);
	    return 0;
	}
    }
    rpair[rpairs].col = col;
    rpair[rpairs].row = row;

    rpairs++;
    rpairs_sorted = 0;	        /* Clobber the flag indicating that the data
                                   in the stroage vector is sorted */

    return 1;			/* success */
}

/**
 * Perform a binary search through the stored set of <row,column> pairs.
 * the routine returns when it finds an ellement in the set which has
 * a matching 'col' and who's row is less than that supplied as the
 * row argument.
 *
 * If neccessary this routine will sort the data in the collection.
 *
 * @param col is the column parameter.
 * @param row is the row parameter.
 *
 * @returns the row component of the selected element.
 */
int
remember(int col, int row)
{
    RPAIR rp;
    int left, right, middle, d;

    if (!rpairs_sorted) {
	qsort(rpair, rpairs, sizeof(RPAIR), rpair_cmp);
	rpairs_sorted = 1;
    }
    rp.col = col;
    rp.row = row;

    left = 0;
    right = rpairs - 1;

    if (rpair_cmp(&rpair[left], &rp) <= 0 && (rpair_cmp(&rpair[right], &rp) >= 0)) {
	while (right - left > 1) {
	    middle = (left + right) / 2;
	    d = rpair_cmp(&rpair[middle], &rp);
	    if (d < 0)
		left = middle;
	    else if (d >= 0)
		right = middle;
	}
	if (rpair_cmp(&rpair[right], &rp) >= 0 && rpair[left].col == col && rpair[left].row < row)	/* success */
	    return rpair[left].row;
    }
    /* error - this should never happen
     */
    fprintf(stderr,
         "ERROR in remember() left: %d (%d %d) right: %d (%d %d) col: %d row: %d\n",
         left, rpair[left].col, rpair[left].row, right, rpair[right].col,
         rpair[right].row, col, row);
    exit(1);
}
