/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

static const char *cvsid = "$Id: rand3.c 10638 2003-04-15 09:15:12Z dgm $";

/**
 * @file 
 *
 * Random number generation algorithim taken from Numerical recipes in C,
 * see http://www.ulib.org/webRoot/Books/Numerical_Recipes/bookcpdf.html
 * There has been a modification by RFM (in 1991) to apparently
 * prevent this routine occasionaly returning a negative number.
 *
 * <b>NB.</b> the routine defined in this file is not <em>exactly</em> as
 * defined in the reference (see original in chapter 7.1, page 283 of
 * the above reference). 
 */


#define MBIG 1000000000
#define MSEED 161803398
#define MZ 0
#define FAC (1.0/MBIG)

/**<
 * @define 
 * According to Knuth, any large MBIG, and any smaller (but still large)
 * MSEED can be substituted or the above values.
 */

/**
 *
 * Knuth's subtractive method for generating random numbers.
 *
 * @param idum is set to a negative value to reinitalise the generator.
 * 
 * @returns a uniform random deviate between 0.0 and 1.0.
 */

float 
rand3(int *idum)
{

    static int inext, inextp;
    static long ma[56];		/* The value 56 (range ma[1..55]) is special
				 * and should not be modified; see Knuth.
				 */

    static int iff = 0;
    long mj, mk;
    int i, ii, k;
    float ZZ;

    if (*idum < 0 || iff == 0) {	/* Initialization. */
	iff = 1;
	mj = MSEED - (*idum < 0 ? -*idum : *idum);	/* VARIATION */
	mj %= MBIG;
	ma[55] = mj;
	mk = 1;
	for (i = 1; i <= 54; i++) {

	    /*
	     * Now initialize the rest of the table, in
	     * a slightly random order, with numbers that
	     * are not especially random.
	     */

	    ii = (21 * i) % 55;
	    ma[ii] = mk;
	    mk = mj - mk;
	    if (mk < MZ)
		mk += MBIG;
	    mj = ma[ii];
	}

	/*
	 * We randomize them by "warming up the generator."
	 */

	for (k = 1; k <= 4; k++)
	    for (i = 1; i <= 55; i++) {
		ma[i] -= ma[1 + (i + 30) % 55];
		if (ma[i] < MZ)
		    ma[i] += MBIG;
	    }
	inext = 0;		/* Prepare indices for our first generated number. */
	inextp = 31;		/* The constant 31 is special; see Knuth. */
	*idum = 1;
    }
    if (++inext == 56)
	inext = 1;		/* Increment inext and inextp, wrapping
				 * around 56 to 1.
				 */
    if (++inextp == 56)
	inextp = 1;
    mj = ma[inext] - ma[inextp];	/* Generate a new random number
					 * subtractively.
					 */
    if (mj < MZ)
	mj += MBIG;		/* Be sure that it is in range. */
    ma[inext] = mj;		/* Store it, */
    /*      return mj*FAC; *//* VARIATION */
    ZZ = mj * FAC;		/* and output the derived uniform deviate. */
    ZZ = (ZZ < 0.0 ? -ZZ : ZZ);
    ZZ = (ZZ > 1.0 ? ZZ - (int) ZZ : ZZ);
    return (ZZ);

}
