/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

/* functions for reading in lines (C) Richard Mott,, ICRF */

static const char *cvsid = "$Id: readline.c 10638 2003-04-15 09:15:12Z dgm $";

/**
 * @file readline.c
 * @brief Utility functions for IO with comment stripping and test fn's
 */

#include <stdio.h>
#include <ctype.h>

#include "readline.h"

/**
 * Truncates the supplied string at the first '!' character. <b>NB.</b> This
 * is a destructive operator which alters the supplied string.
 *
 * @param string is the pointer to the data(the string to be processed).
 */
void 
tts_uncomment(char *string)
{

    while (*string != '!' /*&& *string != '#' */  && *string != 0)
	string++;
    *string = 0;
}


/**
 * Read individual lines from the supplied FILE stream. The line terminator
 * is considered to be the '\n' character. The data is copied into (and over)
 * the supplied string buffer.
 *
 * @param file is the FILE descriptor to use for reading data.
 * @param string is the target buffer to store the string data in.
 *
 * @returns the length of the string read or the value 'EOF' if the end
 * of file was encountered during the read. In the case of the last line
 * in the file the value EOF will be returned (instead of the length) if
 * there is no final newline character in the file.
 */
int 
tts_read_line(FILE * file, char *string)
{
    int c;
    int i = 0;

    while ((c = getc(file)) != EOF) {
	if (c == '\n')
	    break;
	string[i++] = c;
    }
    string[i] = '\0';

    if (c == EOF)
	return i ? i : EOF;

    return i;
}

/**
 * Advances the file pointer until one character after the next new-line
 * character. 
 *
 * @param file is the FILE handle for the required data source.
 *
 * @returns 1 if EOF is encountered during the scan for the next new-line and
 * 0 otherwise.
 */
int 
tts_next_line(FILE * file)
{
    int c;

    while ((c = getc(file)) != EOF) {
	if (c == '\n')
	    return 1;
    }
    return 0;
}

/**
 * Checks whether string is full of white space (that is, it contains no
 * character that would return true for the isspace macro). (<b>NB</b>
 * the routine returns after the first matching character is found).
 *
 * @param string is the pointer to the data (the string to be analysed).
 *
 * @returns 1 if a non-whitespace character is found and 0 otherwise.
 */
int 
tts_not_blank(char *string)
{

    while (*string != 0) {
	if (!isspace((int) *string))
	    return 1;
	string++;
    }
    return 0;
}

/**
 * A wrapped version of read_line. This version deals with removal of comments
 * and blank lines. This routine will return the first (and subsequent) lines
 * for the file that are not commented and not blank.
 *
 * @param fp is the FILE descriptor to use for reading data.
 * @param string is the target buffer to store the string data in.
 * 
 * @see tts_read_line
 *
 * @returns the length of the string read or the value 'EOF' if the end
 * of file was encountered during the read. In the case of the last line
 * in the file the value EOF will be returned (instead of the length) if
 * there is no final newline character in the file.
 */
int 
tts_skip_comments(FILE * fp, char *string)
{

    int n;

    *string = 0;
    while ((n = tts_read_line(fp, string)) != EOF) {
	tts_uncomment(string);
	if (tts_not_blank(string))
	    return n;
    }
    return n;
}
