/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <cl.h>
#include "searchpath.h"

static const char *cvsid = "$Id: searchpath.c 35448 2010-07-01 15:08:05Z jkb $";


/**
 * Extract from the command line the name of a file to open. The user
 * should specify only the basename since this routine ensures that a
 * specific extension is added to than string before opening. 
 *
 * @param format is the string specifying the command line flag that
 *     should be used when determining the base name of the file that
 *     should be opened.
 * @param ext is the extension that should be appended to the base name
 *     before atempting to open the file.
 * @param mode is the file opening mode that should be passed to fopen.
 * @param argc is the count of arguments on the command line.
 * @param argv is the vector of command line arguments.
 * @param fullname is the actual name of the file that has been opened.
 *
 * @returns the file stream associated with the opened file.
 */
FILE *
openfile_in_seqpath_arg(char *format,
			char *ext, char *mode, int argc,
			char **argv, char *fullname)
{
    char buf[256];

    if (getarg(format, buf, argc, argv))
	return openfile_in_seqpath(buf, ext, mode, fullname);
    else
	return NULL;
}

/**
 * Combine a file name with an extension and endevour to open it by
 * searching on a path.
 *
 * @param basename is the unextended basic form of the required file name.
 * @param ext is the extension that should be added to the basename before
 *     atempting to open it.
 * @param mode is the mode that should be passed to fopen
 * @param fullname is a return parameter that is bound to the name of the
 *     file that is actually opened.
 *
 * @returns the file stream associated with the opened file.
 */
FILE *
openfile_in_seqpath(char *basename, char *ext, char *mode, char *fullname)
{
    char buf[256];

    strcpy(buf, basename);
    extension(buf, ext);

    return openfile_in_envpath(buf, mode, SEQPATH, fullname);
}

/**
 * Find and open a file by scanning the supplied search path.
 *
 * @param basename is the name of the file without directory
 * @param mode is the file open mode <em>cf.</em> fopen
 * @param searchpath is the colon-separated search path.
 * @param fullname is a return parameter indicating the full name of the file
 *        that was successfully opened. (or NULL in the event of error).
 *
 * @returns a file pointer to the file with valid operations being defined
 *        by the value of the 'mode' argument.
 */
FILE *
openfile_in_searchpath(char *basename,
		       char *mode,
		       char *searchpath,
		       char *fullname)
{
    FILE *fp = NULL;

    /* if search path defined... */
    if (searchpath) {
	char *s = strdup(searchpath);
	char *p = strtok(s, ":");
	while (p) {
	    sprintf(fullname, "%s/%s", p, basename);
	    if ((fp = fopen(fullname, mode)) != 0)
		break;
	    p = strtok(NULL, ":");
	}
	free(s);
    }
    /* try to open in cwd */
    if (!fp) {
	strcpy(fullname, basename);
	fp = fopen(basename, mode);
    }
    if (!fp)
	*fullname = NULL;

    return fp;
}

/**
 * Open a file in a search path specified by the environment variable env
 *
 * @param basename is the name of the file without directory
 * @param mode is the file open mode <em>cf.</em> fopen
 * @param env is the name of the environment variable to be used when determining
 *        the search path. The value of the environment variable is used as the
 *        search path.
 * @param fullname is a return parameter indicating the full name of the file
 *        that was successfully opened. (or NULL in the event of error).
 *
 * @returns a file pointer to the file with valid operations being defined
 *        by the value of the 'mode' argument.
 */
FILE *
openfile_in_envpath(char *basename, char *mode, char *env, char *fullname)
{
    return openfile_in_searchpath(basename, mode, getenv(env), fullname);
}
